 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: interface.c
 *
 * Description: Server module for plug-in interface.
 *
 *
 * Global Functions: See HTML documentation
 * Global Variables: See HTML documentation
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include "rover.h"
#include "rovertcp.h"
#include "RoverEscape.h"

/* #define TIMER */


PRIVATE void printStats(char *label);
PUBLIC int statsAdd(double value);
PRIVATE void statsClear();
PRIVATE int statsInit(int count);
PRIVATE double *samples = NULL;

PRIVATE char line[2050], command[1024];

typedef void (*plugIn)(char *object, char *objclass);
typedef void (*plugRec)(CMDtype cmd);
typedef void (*plugLock)(CMDtype cmd);

struct _PlugIns {
  char *id;		/* The module's namespace prefix id */
  plugIn import;	/* The module's import procedure */
  plugIn export;	/* The module's export procedure */
  plugRec recover;	/* The module's recovery procedure */  
  plugIn command;	/* The module's command procedure */
};
typedef struct _PlugIns PlugIns;

/* Server Plug-In Modules */
#ifdef USE_FULL
extern void email_import(char *object, char *objclass);
extern void email_export(char *object, char *objclass);
extern void ical_export(char *object, char *objclass);
extern void ical_import(char *object, char *objclass);
extern void nntp_import(char *object, char *objclass);
extern void nntp_export(char *object, char *objclass);
extern void irolo_import(char *object, char *objclass);
extern void irolo_export(char *object, char *objclass);
extern void irolo_command(char *object, char *objclass);
extern void agent_import(char *object, char *objclass);
extern void agent_export(char *object, char *objclass);
extern void agent_recover(CMDtype cmd);
extern void rfs_import(char *object, char *objclass);
extern void rfs_export(char *object, char *objclass);
#endif /* USE_FULL */

#ifdef USE_WEBCAL
extern void webcal_export(char *object, char *objclass);
extern void webcal_import(char *object, char *objclass);
#endif /* USE_WEBCAL */


PRIVATE PlugIns plugs[] = {
#ifdef USE_FULL
  { "Email__", email_import, email_export, NULL, NULL },
  { "Ical__", ical_import, ical_export, NULL, NULL },
  { "Nntp__",	nntp_import, nntp_export, NULL, NULL },
  { "Irolo__", irolo_import, irolo_export, NULL, irolo_command },
  { "Agent__", agent_import, agent_export, agent_recover, NULL },
  { "RFS__", rfs_import, rfs_export, NULL, NULL },
#endif /* USE_FULL */
#ifdef USE_WEBCAL
  { "Webcal__", webcal_import, webcal_export, NULL, NULL },
#endif /* USE_WEBCAL */
  /* Add new plug-in modules here */
  { NULL, NULL, NULL, NULL, NULL }
};

PUBLIC void perform_command(ServerState *s)
{
  CMDtype cmd;
  char *value, *class, *type;
  struct stat sbuf;
  FILE *file;
  int size, i, rc, extra;

  if (s->auth.Nobody) {
    s->status = 500;
    errorExit("Server reached protected procedure");
  }

  /* Do the command */
  if (!s->command) {
    s->status = 400;
    errorExit("No command (CMD) string found");
  }
  cmd = Name2Cmd(s->command);

  switch(cmd) {
  case TEST:
    if (!s->size) {
      s->status = 400;
      errorExit("No size (SIZE) string found");
    }
    testExit(atoi(s->Object));
    break;

  case HTTP:
    gettimeofday(&AppStartTv, &Tz);
    processHTTP(s, 0, NULL, 0);
    gettimeofday(&AppEndTv, &Tz);
    fprintf(stderr, "HTTP time is %f\n", elapsedTime(AppStartTv, AppEndTv));
    statsAdd(elapsedTime(AppStartTv, AppEndTv));
    break;

  case RESET:
    /* If there are any other open connections, this will silently fail */
    s->ObjType = ROVER_CMD_COMPLETE;
    fprintf(stderr, "Got reset command: %s\n", s->Object);
    if (!strncmp("ARP_Reset", s->Object, strlen("ARP_Reset"))) {
      /* Request for ARP cache reset: used when the client switches networks */
      if (s->keepalive) {
	s->status = 500;
	errorExit("Got RESET in the middle of a batched request set\n");
      }
      rc = sendServerResponse(&extra, 0);
      s->CommandDone = 1;
      close(s->tcp->sock);
      fclose(s->Stdin);
      fclose(s->Stdout);
      s->Stdin = s->Stdout = NULL;
      sleep(2);
      reset_arp(s, s->rhost);
      if (strcmp(s->rhost, s->httpdHost)) reset_arp(s, s->httpdHost);
    } else if (!strncmp("LOG__", s->Object, strlen("LOG__"))) {
      /* Request for change in server logging actions */
      if (s->keepalive) {
	s->status = 500;
	errorExit("Got RESET in the middle of a batched request set\n");
      }
      class = index(s->Object, '_');
      class += 2;
      LogChange(class);
      rc = sendServerResponse(&extra, 0);
      s->CommandDone = 1;
    } else if(!strncmp("MSG__", s->Object, strlen("MSG__"))) {
      class = index(s->Object, '_');
      class += 2;
      fprintf(stderr, "Client Message: %s\n", class);
#ifdef TIMER
      if (TotalTimeCount)
	fprintf(stderr,
		"Client Iters %d Avg App Time %.4f Avg Server Time %.4f\n",
		TotalTimeCount, TotalAppTime/((double)TotalTimeCount),
		TotalServerTime/((double)TotalTimeCount));
      TotalTimeCount = 0;
      TotalServerTime = TotalAppTime = 0.0;
#endif
      if (samples) {
	printStats("Server-side Stats");
	statsClear();
      }
      rc = sendServerResponse(&extra, 0);
      s->CommandDone = 1;
    } else if(!strncmp("CODE__", s->Object, strlen("CODE__"))) {
      /* ############### Yes, this is a BIG security hole! ############### */
      class = index(s->Object, '_');
      class += 2;
      system(class);
      rc = sendServerResponse(&extra, 0);
      s->CommandDone = 1;
    } else if(!strncmp("Rover__", s->Object, strlen("Rover__"))) {
      value = copyString(s->Object);
      class = index(value, '_');
      class += 2;
      
      if (RoverDebug)
	fprintf(stderr, "Command request for |%s|\n", class);
      i = 0;
      while ((plugs[i].id != NULL) &&
	     (strncmp(plugs[i].id, class, strlen(plugs[i].id))))
	i++;
      if (plugs[i].id != NULL) {
#ifdef TIMER
	gettimeofday(&AppStartTv, &Tz);
#endif TIMER
	(plugs[i].command)(value, class);
#ifdef TIMER
	gettimeofday(&AppEndTv, &Tz);
#endif TIMER
      } else {
	sprintf(line, "Unknown class (%s)\n", class);
	errorExit(line);
      }
      free(value);
    } else {
      fprintf(stderr, "Unknown RESET command: %s\n", s->Object);
      rc = sendServerResponse(&extra, 0);
      s->CommandDone = 1;
    }
    break;

  case IMPORTDATA:
    /* Ordinary object */
    s->ObjType = ROVER_TCL_TK;
    
    if(!strncmp("Rover__tclIndex__", s->Object,
		strlen("Rover__tclIndex__"))) {
      /* Request for Rover__tclIndex object */
      tcl_importdata(s->Object);
    } else {
      strcpy(command, RoverRoot[RoverRootID]);
      strcat(command, "/");
      strcat(command, s->Object);
      
      if (!stat(command, &sbuf)) {
	if ((sbuf.st_mode & (S_IWUSR|S_IWGRP|S_IWOTH)) == 0) {
	  fprintf(stderr, "Read-only object ==> ROVER_CACHE_FOREVER\n");
	  s->Cacheable = ROVER_CACHE_FOREVER;
	  s->UserTime = (time_t) 0;
	}
	s->dvtime = sbuf.st_mtime;
	if (checkVerify(s->Object, s, ROVER_TCL_TK_UPDATE)) return;
	outstrResize(sbuf.st_size+10);
      } else {
	fprintf(stderr, "No object found (%s) [interface.c]!\n", command);
	s->status = 404;
	errorExit("No object found");
      }
      
      file = fopen(command,"r");
      if (file == NULL) {
	perror("File open failed\n");
	s->status = 404;
	errorExit("File open failed");
      }
      emitData();
      while (1) {
	line[0] = '\0';
	rc = fread(line, sizeof(char), 1023, file);
	line[rc] = '\0';
	outstrD(line);
	if (feof(file)) break;
      }
      fclose(file);
    }
    break;

  case IMPORTCMD:
    /* Check if the object is a null qrpc */
    if (!strncmp("QRPC", s->Object, strlen("QRPC"))) {
      s->ObjType = ROVER_TCL_TK;
      s->Cacheable = ROVER_NO_CACHE;
 
      /* Check if the object is a timer object */
    } else if (!strncmp("Time__", s->Object, strlen("Time__"))) {
      s->ObjType = ROVER_TCL_TK;
      value = copyString(s->Object);
      type = index(value, '_');
      type += 2;
      if (!strncmp("NC__", type, strlen("NC__"))) {
	s->Cacheable = ROVER_NO_CACHE;
      } else {
	s->Cacheable = ROVER_CACHE_FOREVER;
      }
      class = index(type, '_');
      *class = '\0';
      class += 2;
      size = atoi(class);
      outstrVC("proc Time__%0__%1 {} {}\nreturn \"nullCallback\"\n", 2,
	       type, class);
      fprintf(stderr, "Making timing object w/ %d code and %d data bytes\n",
	      s->CodeSize+1, size);
/*    size -= (s->CodeSize+2);
      if (size < 0) return; */
      /* Now we make an object */
      emitData();
      /* Adjust the buffer's size */
      outstrResize(size);
      while (size--)
	outstrD("X");
      free(value);

      /* Check if the object is a Rover object */
    } else if (strncmp("Rover__", s->Object, strlen("Rover__"))) {
      /* Nope, must be an ordinary object */
      s->ObjType = ROVER_TCL_TK;

      strcpy(command, RoverRoot[RoverRootID]);
      strcat(command, "/");
      strcat(command, s->Object);

      if (!stat(command, &sbuf)) {
	if ((sbuf.st_mode & (S_IWUSR|S_IWGRP|S_IWOTH)) == 0) {
	  fprintf(stderr, "Read-only object ==> ROVER_CACHE_FOREVER\n");
	  s->Cacheable = ROVER_CACHE_FOREVER;
	  s->UserTime = (time_t) 0;
	}
	s->dvtime = sbuf.st_mtime;
	if (checkVerify(s->Object, s, ROVER_TCL_TK_UPDATE)) return;
	outstrResize(sbuf.st_size+10);
      } else {
	fprintf(stderr, "No object found (%s)!\n", command);
	s->status = 404;
	errorExit("No object found");
      }

      file = fopen(command,"r");
      if (file == NULL) {
	perror("File open failed\n");
	s->status = 404;
	errorExit("File open failed");
      }
      while (1) {
	line[0] = '\0';
	rc = fread(line, sizeof(char), 1023, file);
	line[rc] = '\0';
	outstrC(line);
	if (feof(file)) break;
      }
      fclose(file);
    } else {
      value = copyString(s->Object);
      class = index(value, '_');
      class += 2;
      class = copyString(class);
      
      if (RoverDebug)
	fprintf(stderr, "Import request for |%s|\n", class);
      i = 0;
      while ((plugs[i].id != NULL) &&
	     (strncmp(plugs[i].id, class, strlen(plugs[i].id))))
	i++;
      if (plugs[i].id != NULL) {
	gettimeofday(&AppStartTv, &Tz);
	(plugs[i].import)(value, class);
	gettimeofday(&AppEndTv, &Tz);
      } else {
	sprintf(line, "Unknown class (%s)\n", class);
	errorExit(line);
      }
      free(value);
      free(class);
    }
    break;

  case EXPORTCMD:
    /* Check if the object is a timer object */
    if (!strncmp("Time__", s->Object, strlen("Time__"))) {
      if (createTclInterp(state) < 0)  {
	fprintf(stderr, "Server unable to create tcl interpreter\n");
	s->status = 500;
	errorExit("Server unable to create tcl interpreter");
      }
      s->ObjType = ROVER_TCL_TK;
      value = copyString(s->Object);
      type = index(value, '_');
      type += 2;
      s->Cacheable = ROVER_NO_CACHE;
      class = index(type, '_');
      *class = '\0';
      class += 2;
      outstrVC("proc Time__F__%1 {} {}\nreturn \"nullCallback\"\n", 1, class);
      fprintf(stderr, "Timing object export: %d data bytes\n",
	      s->CodeSize+1);
      deleteTclInterp();
      free(value);

      /* Check if the object is a Rover object */
    } else if (strncmp("Rover__", s->Object, strlen("Rover__"))) {
      /* Nope, must be an ordinary object */
      errorExit("Can't export ordinary obj [perform_command:interface.c]\n");
    } else {
      value = copyString(s->Object);
      class = index(value, '_');
      class += 2;
      
      if (RoverDebug)
	fprintf(stderr, "Export request for |%s|\n", class);
      i = 0;
      while ((plugs[i].id != NULL) &&
	     (strncmp(plugs[i].id, class, strlen(plugs[i].id))))
	i++;
      if (plugs[i].id != NULL) {
	gettimeofday(&AppStartTv, &Tz);
	(plugs[i].export)(value, class);
	gettimeofday(&AppEndTv, &Tz);
	fprintf(stderr,"Export time is %f\n",elapsedTime(AppStartTv,AppEndTv));
	statsAdd(elapsedTime(AppStartTv, AppEndTv));
      } else {
	sprintf(line, "Unknown class (%s)\n", class);
	errorExit(line);
      }
      free(value);
    }

    break;
  default:
    sprintf(line, "Unknown command (%s)\n", s->command);
    errorExit(line);
    break;
  }
  if (s->Stdout) fflush(s->Stdout);
}

PUBLIC void recover_command(ServerState *s)
{
  CMDtype cmd;
  char *value = NULL, *class;
  int i;

  if (s->auth.Nobody) {
    s->status = 500;
    errorExit("Server reached protected procedure");
  }

  /* Do the command */
  if (!s->command) {
    s->status = 400;
    errorExit("No command (CMD) string found");
  }
  cmd = Name2Cmd(s->command);

  switch(cmd) {
  case TEST:
    /* Simply rerun the command */
    perform_command(s);
    break;

  case IMPORTDATA:
    /* Ordinary object, simply rerun the command */
    perform_command(s);
    break;

  case IMPORTCMD:
    /* Check if the object is a null qrpc */
    if (!strncmp("QRPC", s->Object, strlen("QRPC"))) {
      /* Simply rerun the command */
      perform_command(s);
 
      /* Check if the object is a timer object */
    } else if (!strncmp("Time__", s->Object, strlen("Time__"))) {
      /* Simply rerun the command */
      perform_command(s);

      /* Check if the object is a Rover object */
    } else if (strncmp("Rover__", s->Object, strlen("Rover__"))) {
      /* Nope, must be an ordinary object, simply rerun the command */
      perform_command(s);
    } else {
      value = copyString(s->Object);
      class = index(value, '_');
      class += 2;
      
      if (RoverDebug)
	fprintf(stderr, "Rerunning Import request for |%s|\n", class);
      i = 0;
      while ((plugs[i].id != NULL) &&
	     (strncmp(plugs[i].id, class, strlen(plugs[i].id))))
	i++;
      if (plugs[i].id != NULL) {
	/* 
	 * The Plug In C recovery procedure should do the following:
	 *   1. Restore the C/Tcl environment.
	 *   2. Restore the stable variables by calling restoreStableVars()
	 *   3. Invoke the Tcl recovery procedure specified by
	 *        s->recoveryProc.
	 */
	if (plugs[i].recover)
	  (plugs[i].recover)(cmd);
	else {
	  sprintf(line, "Rover class (%s) doesn't have a C recover command\n",
		  class);
	  errorExit(line);
	}
      } else {
	sprintf(line, "Unknown class (%s)\n", class);
	errorExit(line);
      }
      free(value);
    }
    break;

  case EXPORTCMD:
    /* Check if the object is a timer object */
    if (!strncmp("Time__", s->Object, strlen("Time__"))) {
      /* Simply rerun the command */
      perform_command(s);

      /* Check if the object is a Rover object */
    } else if (strncmp("Rover__", s->Object, strlen("Rover__"))) {
      /* Nope, must be an ordinary object */
      errorExit("Can't export ordinary obj [perform_command:interface.c]\n");
    } else {
      value = copyString(s->Object);
      class = index(value, '_');
      class += 2;
      
      if (RoverDebug)
	fprintf(stderr, "Export request for |%s|\n", class);
      i = 0;
      while ((plugs[i].id != NULL) &&
	     (strncmp(plugs[i].id, class, strlen(plugs[i].id))))
	i++;
      if (plugs[i].id != NULL) {
	/* 
	 * The Plug In C recovery procedure should do the following:
	 *   1. Restore the C/Tcl environment.
	 *   2. Restore the stable variables by calling restoreStableVars()
	 *   3. Invoke the Tcl recovery procedure specified by
	 *        s->recoveryProc.
	 */
	if (plugs[i].recover)
	  (plugs[i].recover)(cmd);
	else {
	  sprintf(line, "Rover class (%s) doesn't have a C recover command\n",
		  class);
	  errorExit(line);
	}
      } else {
	sprintf(line, "Unknown class (%s)\n", class);
	errorExit(line);
      }
      free(value);
    }

    break;
  default:
    sprintf(line, "Unknown command (%s)", s->command);
    errorExit(line);
    break;
  }
  if (s->Stdout) fflush(s->Stdout);
}


#include <stdio.h>
#include <float.h>
#include <math.h>

int sampleSize = 0;
int sampleMax = 0;

#define DEFAULTSIZE	32

PRIVATE int statsInit(int count)
{
  if (count)
    sampleMax = count;
  else
    sampleMax = DEFAULTSIZE;

  sampleSize = 0;
  if ((samples = (double *)malloc(sampleMax*sizeof(double))) == NULL) {
    fprintf(stderr, "Out of memory");
    return(0);
  }
  return(1);
}

PRIVATE void statsClear()
{
  if (samples) {
    free(samples);
    sampleSize = 0;
    samples = NULL;
  }
}

PUBLIC int statsAdd(double value)
{
  if (!samples) statsInit(0);
  if ((sampleSize + 1) >= sampleMax) {
    sampleMax <<= 1;
    if ((samples =(double *)realloc(samples,sampleMax*sizeof(double)))==NULL) {
      fprintf(stderr, "Out of memory");
      return(0);
    }
  }
  samples[sampleSize++] = value;
  return(1);
}

PRIVATE void printStats(char *label)
{
  int count = 0;
  double min = DBL_MAX;
  double max = DBL_MIN;
  double mean = 0.0;
  double sum = 0.0;
  double sumofsq = 0.0;
  double sigma = 0.0;
  double term1, term2;

  if (!samples || !sampleSize) return;
  fprintf(stderr, "n\tx\t\tmean\t\tsigma\n");
  while (count < sampleSize) {
    double x = samples[count];
    if (x < min) min = x;
    if (x > max) max = x;
    sum += x;
    sumofsq += pow(x,2);
    count++;
    if (count == 1) 
      mean = x;
    else {
      term1 = sigma + pow((mean - x), 2) / pow(count,2);
      mean = sum / count;
      term2 = pow((x - mean),2) / count;
      sigma = (count - 1.0) / count * term1 + term2;
    }

    fprintf(stderr,  "%d\t%f\t%f\t%f\n", count, x, mean, sqrt(sigma));
  }
  sigma =  sqrt((sumofsq / count) - pow(mean,2));
  fprintf(stderr,
	  "%s:totals:num=%d:min=%f:max=%f:mean=%f:sigma=%f\n",
	  label, count, min, max, mean, sigma);
  sampleSize = 0;
}
