 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: recover.c
 *
 * Description: Rover Server recovery module.
 *
 *
 * Global Functions: 
 *	int checkMsgID(char *user, char *host, int msgID, unsigned int incarnation, int retry);
 *	void saveResult(int seq, int rc);
 *	void saveState(ServerState *s);
 *	void recover(char *name, ServerState *s);
 *	void LogChange(char *action);
 *	int cleanLog();
 *
 * Global Variables: 
 *	int logClean = 1;
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1996-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include "rover.h"
#include "rovertcp.h"
#include "html.h"

#define DEBUG_RECOVER

PUBLIC int logClean = 1;

typedef enum _CheckType {
  CHKEOF = -1,
  CHKOK = 0,
  CHKERR = 1
} CheckType;

PRIVATE int ServerSequence = 0;
PRIVATE int execOrder = 0;
PRIVATE char command[256];

/* Delete all messages from an earlier incarnation */
PRIVATE void flushMsgs(char *user, char *host, int msgID, unsigned int incarnation)
{
  int i;

  /* Flush any pending messages from an earlier incarnation of the series */
  for (i = 0; i < MAXSTATES; i++)
    if (States[i].inuse && (States[i].reqType != NOTYPE) &&
	!States[i].pid && States[i].msgID && !strcmp(States[i].ClientUser, user) &&
	!strcmp(States[i].rhost, host) && (States[i].incarnation < incarnation)) {
      /* Delete the message */
      fprintf(stderr, ">>>Flushing msgID %d (inc: %u) from %s@%s\n", States[i].msgID,
	      States[i].incarnation, user, host);
      if (States[i].Stdin) {
	if (!States[i].keepalive && (States[i].tcp->sock != -1)) {
	  close(States[i].tcp->sock);
	  States[i].tcp->sock = -1;
	  fclose(States[i].Stdin);
	  fclose(States[i].Stdout);
	}
	States[i].Stdin = States[i].Stdout = NULL;
      }
      reset_state(&(States[i]));
    }
  sprintf(command, "MsgID_%s@%s", user, host);
  fprintf(ServerLogFile, "ROVER_LOG_MSGID %s %d %u\n", command, msgID+1, incarnation);
  fflush(ServerLogFile);
  if (ServerHardFlush) fsync(ServerLogFD);
}

/* 
 * Check if a message ID is acceptable and update the expected message ID.
 *   Returns 1 if the message is acceptable; if the message should be deleted,
 *   returns -1; otherwise, returns 0.
 */
PUBLIC int checkMsgID(char *user, char *host, int msgID, unsigned int incarnation, int retry)
{
  int new;
  ClientMsgID *cm = NULL;
  
  if (retry) fprintf(stderr, "MsgID %d has incarnation = %u, retry count = %d\n",
		     msgID, incarnation, retry);
  sprintf(command, "MsgID_%s@%s", user, host);
  hentry = Tcl_FindHashEntry(&Info, command);
  if (hentry && ((cm = (ClientMsgID *) Tcl_GetHashValue(hentry)) != NULL)) {
    /* The server expects either:
       1. A msgID of zero (0) indicating a new message series.
       2. A msgID equal to the expected msgID.
       3. A msg from a later incarnation.
     */
    if (incarnation != cm->incarnation) {	/* Not the expected incarnation */
      if (incarnation < cm->incarnation) {	/* Message is from an old incarnation */
	fprintf(stderr, ">>>Old msgID from %s@%s: %d (inc: %u), expected %d (inc: %u)\n",
		user, host, msgID, incarnation, cm->nextID, cm->incarnation);
	return(-1);
      } else {				/* Message is from new incarnation */
	if (msgID)
	  fprintf(stderr,
		  ">>>WARNING! No message series reset from %s@%s for new message incarnation: %d (inc: %u-> %u)\n",
		  user, host, msgID, cm->incarnation, incarnation);
	fprintf(stderr, ">>>Flushing messages for series %s@%s\n", user, host);
	flushMsgs(user, host, msgID, incarnation);
	cm->nextID = msgID+1;		/* Update the expected msgID */
	cm->incarnation = incarnation;
      }
      return(1);
    }

    /* Expected incarnation, process as before */
    if (msgID && (msgID != cm->nextID)) {	/* Not the expected msgID */
      if (msgID < cm->nextID) {
	fprintf(stderr, ">>>Duplicate msgID from %s@%s: %d (inc: %u), expected %d (inc: %u)\n",
		user, host, msgID, incarnation, cm->nextID, cm->incarnation);
	return(-1);
      } 
      fprintf(stderr,
	      ">>>Early (out of order) msgID from %s@%s: %d (inc: %u), expected %d (inc: %u)\n",
	      user, host, msgID, incarnation, cm->nextID, cm->incarnation);
      return(0);
    } else if (!msgID) {		/* Reset the message series */
      fprintf(stderr, ">>>Message series reset from %s@%s: %d (inc: %u), expected %d (inc: %u)\n",
	      user, host, msgID, incarnation, cm->nextID, cm->incarnation);
      flushMsgs(user, host, msgID, incarnation);
    }
    cm->nextID = msgID+1;		/* Update the expected msgID */

  } else {				/* New message series */
    if (msgID)
      fprintf(stderr,
	      ">>>WARNING! No message series reset from %s@%s for new message series: %d (inc: %u)\n",
	      user, host, msgID, incarnation);
    hentry = Tcl_CreateHashEntry(&Info, command, &new);
    if (!new) {				/* Check for duplicate sequences */
      fprintf(stderr, "Duplicate Info hash table entry: %s [CheckMsgID:recover.c]\n", 
	      command);
      exit(-1);
    }
    if ((cm = (ClientMsgID *)malloc(sizeof(ClientMsgID))) == NULL) return(CHKERR);
    cm->nextID = msgID+1;		/* Update the expected msgID */
    cm->incarnation = incarnation;
    Tcl_SetHashValue(hentry, (char *)cm);	/* Save it */
  }  
  return(1);
}

PUBLIC void saveResult(int seq, int rc)
{
  if (!ServerNoLog) {
    fprintf(ServerLogFile, "ROVER_LOG_RESULT %d %d\n", seq, rc);
    fflush(ServerLogFile);
    /* No need for a hard flush: either the next request or
     *   the OS will flush the information
     */
  }
}  

PUBLIC void saveState(ServerState *s)
{
  int rc, csize, psize, osize;
  
  if (!s->inuse) return;
  if (RoverDebug) fprintf(stderr, "Saving state for msgid %d\n", s->msgID);

  s->ServerSeq = ++ServerSequence;
  if (ServerNoLog) return;
  if (s->code)
    csize = strlen(s->code)+1;
  else 
    csize = 0;
  if (s->password)
    psize = strlen(s->password)+1;
  else
    psize = 0;
  osize = strlen(s->Object)+1;
  fprintf(ServerLogFile,
	  "ROVER_LOG_DATA %d %s %s %d %u %s %d %s %d %d %lu %d %d %s %d %s %d %d %d %d %d\n",
	  s->ServerSeq, s->rhost, s->ClientUser, s->msgID, s->incarnation, 
	  s->httpdHost, s->port, s->Service, s->ServiceID, s->LogID,
	  (unsigned long) s->UserTime, s->UserServer, osize, s->command,
	  s->size, s->ContentType, s->ContentLength, s->retry, s->compress,
	  csize, psize);
  
  if ((rc = fwrite(s->Object,sizeof(char),osize,ServerLogFile)) != osize) {
    state->status = 500;
    sprintf(command, "Bad object name write size %d [saveState:recover.c]\n",rc);
    errorExit(command);
  }
  if (csize)
    if ((rc = fwrite(s->code,sizeof(char),csize,ServerLogFile)) != csize) {
      state->status = 500;
      sprintf(command, "Bad code write size %d [saveState:recover.c]\n", rc);
      errorExit(command);
    }
  if (psize)
    if ((rc = fwrite(s->password,sizeof(char),psize,ServerLogFile)) != psize) {
      state->status = 500;
      sprintf(command, "Bad password write size %d [saveState:recover.c]\n", rc);
      errorExit(command);
    }
  fputc('\n', ServerLogFile);
  fflush(ServerLogFile);
  if (ServerHardFlush) fsync(ServerLogFD);
}


/*
 * Check that the number of items read matched the expected number.
 *   Returns CHKOK if the counts matched. Otherwise, returns CHKEOF, if EOF
 *   occurred, and returns CHKERR if the counts did not match.
 */
PRIVATE CheckType rcvCheck(int count, int expect, FILE *file, char *text)
{
  if (count != expect) {
    if (feof(file)) return(CHKEOF);
    fprintf(stderr, "Bad server log %s read %d/%d [rcvCheck:recover.c]\n", text,
	    count, expect);
    if (count == 0)
      return(CHKEOF);
    else
      return(CHKERR);
  }
  return(CHKOK);
}


PRIVATE ServerState *checkState(ServerState *s, int seq)
{
  int i;
  
  if (!s || (s->ServerSeq != seq)) {
    /* Find the right state */
    for (s = NULL, i = 0; i < MAXSTATES; i++) {
      if ((States[i].inuse) && (States[i].ServerSeq == seq))
	break;
    }
    if (i != MAXSTATES) 
      s = &(States[i]);
  }
  return(s);
}


PRIVATE inline LOGtype getLOGtype(char *c)
{
  if (!strcasecmp("ROVER_LOG_VERSION", c))
    return(ROVER_LOG_VERSION);
  else if (!strcasecmp("ROVER_LOG_START", c))
    return(ROVER_LOG_START);
  else if (!strcasecmp("ROVER_LOG_FORK", c))
    return(ROVER_LOG_FORK);
  else if (!strcasecmp("ROVER_LOG_RESULT", c))
    return(ROVER_LOG_RESULT);
  else if (!strcasecmp("ROVER_LOG_RECOVER", c))
    return(ROVER_LOG_RECOVER);
  else if (!strcasecmp("ROVER_LOG_STABLE_VAR", c))
    return(ROVER_LOG_STABLE_VAR);
  else if (!strcasecmp("ROVER_LOG_STABLE_DCL", c))
    return(ROVER_LOG_STABLE_DCL);
  else if (!strcasecmp("ROVER_LOG_STABLE_HTDCL", c))
    return(ROVER_LOG_STABLE_HTDCL);
  else if (!strcasecmp("ROVER_LOG_STABLE_HTVAR", c))
    return(ROVER_LOG_STABLE_HTVAR);
  else if (!strcasecmp("ROVER_LOG_DATA", c))
    return(ROVER_LOG_DATA);
  else if (!strcasecmp("ROVER_LOG_OUTPUT", c))
    return(ROVER_LOG_OUTPUT);
  else if (!strcasecmp("ROVER_LOG_MSGID", c))
    return(ROVER_LOG_MSGID);
  else return(ROVER_LOG_ERROR);
}

PRIVATE CheckType checkLogVersion(FILE *file, ServerState **s)
{
  int rc, value;
  CheckType check;
  
  /* Version info */
  rc = fscanf(file, "%d\n", &value);
  check = rcvCheck(rc, 1, file, "version");
  if (check != CHKOK) return(check);
  if (value != ROVER_SERVER_LOG_VERSION)
    fprintf(stderr,
	    "Unexpected server log version %d (expected %d) [checkLogVersion:recover.c]\n",
	    value, ROVER_SERVER_LOG_VERSION);
  else if (RoverDebug)
    fprintf(stderr, "Got server log version %d\n",value);
  return(CHKOK);
}

PRIVATE inline void setSequenceNumber(int seq)
{
  if (seq > ServerSequence) ServerSequence = seq;
}

PRIVATE CheckType checkLogStart(FILE *file, ServerState **s)
{
  CheckType check;
  int rc, seq;

  /* Server start of execution of a request */
  rc = fscanf(file, "%d\n", &seq);
  setSequenceNumber(seq);
  check = rcvCheck(rc, 1, file, "start of request execution");
  if (check != CHKOK) return(check);
  if (!(*s = checkState(*s, seq))) {
    fprintf(stderr,
	    "[%d] Got LOG_START before LOG_DATA [checkLogStart:recover.c]\n",
	    seq);
    return(CHKERR);
  }
  (*s)->order = ++execOrder;
  if (RoverDebug)
    fprintf(stderr,
	    "[%d] Got server log start of req #%d: %s@%s:%d (inc: %u)\n", seq,
	    (*s)->order, (*s)->ClientUser, (*s)->rhost, (*s)->msgID,
	    (*s)->incarnation);
  return(CHKOK);
}

PRIVATE CheckType checkLogFork(FILE *file, ServerState **s)
{
  CheckType check;
  int rc, seq, pid;
  struct stat sbuf;

  /* Server fork */
  rc = fscanf(file, "%d %d\n", &seq, &pid);
  setSequenceNumber(seq);
  check = rcvCheck(rc, 2, file, "fork");
  if (check != CHKOK) return(check);
  if (RoverDebug)
    fprintf(stderr, "[%d] Got server log fork : %s@%s:%d (inc: %u)\n",
	    seq, (*s)->ClientUser, (*s)->rhost, (*s)->msgID,
	    (*s)->incarnation);
  sprintf(command, "%s/rover-log-file__%s__%s__%d__%u", ServerLogDir,
	  (*s)->ClientUser, (*s)->rhost, (*s)->msgID, (*s)->incarnation);
  if (!stat(command, &sbuf)) {
    recover(command, *s);
  } else {
    fprintf(stderr, "[%d] File %s not found [checkLogFork:recover.c]\n", seq,
	    command);
  }
  return(CHKOK);
}

PRIVATE char *resultText(int rc)
{
  switch (rc) {
  case 0:
    return("ROVER_OK");
    break;
  case ROVER_ERROR:
    return("ROVER_ERROR");
    break;
  case ROVER_FATAL:
    return("ROVER_FATAL");
    break;
  case ROVER_FORK:
    return("ROVER_FORK");
    break;
  case ROVER_PIPE:
    return("ROVER_PIPE");
    break;
  case ROVER_SIGNAL:
    return("ROVER_SIGNAL");
    break;
  }
  return("<unknown>");
}

PRIVATE CheckType checkLogResult(FILE *file, ServerState **s)
{
  CheckType check;
  int rc, seq, value;

  /* Server result */
  rc = fscanf(file, "%d %d\n", &seq, &value);
  setSequenceNumber(seq);
  check = rcvCheck(rc, 2, file, "result");
  if (check != CHKOK) return(check);
  if (!*s) {
    fprintf(stderr, "[%d] Missing ROVER_LOG_DATA [checkLogResult:recover.c]\n",
	    seq);
    return(CHKERR);
  }
  if (RoverDebug)
    fprintf(stderr, "[%d] Got server log result: %s@%s:%d (inc: %u) -> %s (%d)\n",
	    seq, (*s)->ClientUser, (*s)->rhost, (*s)->msgID, (*s)->incarnation,
	    resultText(value), value);

  /* Don't rerun forked or error results */
  if (value && (value != ROVER_FORK) && (value != ROVER_ERROR)) {
    if (RoverDebug)
      fprintf(stderr, "[%d] Non-zero result, request will be re-run\n", seq);
  } else {
    reset_state(*s);
    *s = NULL;
  }
  return(CHKOK);
}

PRIVATE CheckType checkLogRecover(FILE *file, ServerState **s)
{
  CheckType check;
  int rc, seq;
  char tmp[256];

  /* Server recovery procedure */
  rc = fscanf(file, "%d %s\n", &seq, tmp);
  setSequenceNumber(seq);
  check = rcvCheck(rc, 2, file, "recovery procedure");
  if (check != CHKOK) return(check);
  if (!*s) {
    fprintf(stderr, "[%d] Missing ROVER_LOG_DATA [checkLogRecover:recover.c]\n",
	    seq);
    return(CHKERR);
  }
  if ((*s)->recoverProc) free((*s)->recoverProc);
  (*s)->recoverProc = copyString(tmp);
  if (RoverDebug)
    fprintf(stderr, "[%d] Got server log recovery procedure: %s\n", seq, tmp);
  return(CHKOK);
}

PRIVATE CheckType checkLogStableVar(FILE *file, ServerState **s)
{
  CheckType check;
  int found, i, rc, value, seq;
  char c, tmp[256], *buf;

  /* Server stable variable's new value */
  rc = fscanf(file, "%d %s %d", &seq, tmp, &value);
  setSequenceNumber(seq);
  check = rcvCheck(rc, 3, file, "stable variable change");
  if (check != CHKOK) return(check); 
  c = fgetc(file);
  if (RoverDebug) fprintf(stderr, "[%d] Entry size is %d\n", seq, value);
  if (value) {
    if ((buf = (char *)malloc(value)) == NULL) {
      fprintf(stderr, "[%d] Out of memory [checkLogStableVar:recover.c]\n", seq);
      return(CHKERR);
    }
    if ((rc = fread(buf, sizeof(char), value, file)) != value) {
      fprintf(stderr,
	      "[%d] Bad log stable var read size %d [checkLogStableVar:recover.c]\n",
	      seq, rc);
      free(buf);
      return(CHKERR);
    }
  } else {
    buf = malloc(1);
    buf[0] = '\0';
  }
  c = fgetc(file);
  if (!*s) {
    fprintf(stderr,
	    "[%d] Got STABLE_VAR before DATA [checkLogStableVar:recover.c]\n",
	    seq);
    return(CHKERR);
  }
  if ((*s)->ServerSeq != seq) {
    fprintf(stderr,
	    "[%d] STABLE_VAR seq# mis-match [checkLogStableVar:recover.c]\n",
	    seq);
    return(CHKERR);
  }

  for (found = i = 0; i < MAXVAR; i++)
    if ((*s)->vars[i].inuse && !strcmp((*s)->vars[i].name, tmp)) {
      found = 1;
      if ((*s)->vars[i].value) free((*s)->vars[i].value);
      (*s)->vars[i].value = buf;
      break;
    }
  if (RoverDebug) {
    if (found)
      fprintf(stderr,
	      "[%d] Got server stable log var change: %s %s (%d bytes)\n",
	      seq, tmp, buf, value);
    else
      fprintf(stderr,
	      "[%d] Got UNKNOWN server stable log var change: %s %d:%s [checkLogStableVar:recover.c]\n",
	      seq, tmp, value, buf);
  }
  return(CHKOK);
}

PRIVATE CheckType checkLogStableDcl(FILE *file, ServerState **s)
{
  CheckType check;
  int found, i, rc, value, seq;
  char c, tmp[256], *buf;

  /* Server stable var declaration */
  rc = fscanf(file, "%d %s %d", &seq, tmp, &value);
  setSequenceNumber(seq);
  check = rcvCheck(rc, 3, file, "stable variable declaration");
  if (check != CHKOK) return(check); 
  c = fgetc(file);
  if (RoverDebug) fprintf(stderr, "[%d] Entry size is %d\n", seq, value);
  if (value) {
    if ((buf = (char *)malloc(value)) == NULL) {
      fprintf(stderr, "[%d] Out of memory [checkLogStableDcl:recover.c]\n", seq);
      return(CHKERR);
    }
    if ((rc = fread(buf, sizeof(char), value, file)) != value) {
      fprintf(stderr,
	      "[%d] Bad log stable var dcl read size %d [checkLogStableDcl:recover.c]\n",
	      seq, rc);
      free(buf);
      return(CHKERR);
    }
  } else {
    buf = malloc(1);
    buf[0] = '\0';
  }
  c = fgetc(file);
  if (!*s) {
    fprintf(stderr,
	    "[%d] Got STABLE_DCL before DATA [checkLogStableDcl:recover.c]\n",
	    seq);
    free(buf);
    return(CHKERR);
  }
  if ((*s)->ServerSeq != seq) {
    fprintf(stderr,
	    "[%d] STABLE_VAR seq# mis-match [checkLogStableDcl:recover.c]\n",
	    seq);
    free(buf);
    return(CHKERR);
  }

  /* Search for the stable var (to prevent duplicates) and an empty slot */
  for (rc = -1, found = i = 0; i < MAXVAR; i++)
    if ((*s)->vars[i].inuse) {
      if (!strcmp((*s)->vars[i].name, tmp)) {
	found = 1;
	break;
      }
    } else if (rc == -1) rc = i;
  if (!found) {
    /* Not found, good */
    if (rc != -1) {
      /* Got an empty slot, also good */
      i = rc;
      (*s)->vars[i].inuse = 1;
      (*s)->vars[i].name = copyString(tmp);
      (*s)->vars[i].value = buf;
    } else {
      /* No slots, bad */
      fprintf(stderr,
	      "[%d] Out of stable variable slots [checkLogStableDcl:recover.c]\n",
	      seq);
      free(buf);
      return(CHKERR);
    }
  } else {
    /* Found dcl, bad? */
    fprintf(stderr,
	    "[%d] Got DUPLICATE server stable log var dcl: %s %d:%s [checkLogStableDcl:recover.c]\n",
	    seq, tmp, value, buf);
    if (!strcmp((*s)->vars[i].value, buf))
      fprintf(stderr, "[%d] Identical, ignoring\n", seq);
    else {
      free(buf);
      return(CHKERR);
    }
    free(buf);
  }
  return(CHKOK);
}

PRIVATE CheckType checkLogStableHTVar(FILE *file, ServerState **s)
{
  CheckType check;
  int rc, value;
  char c, name[256], entry[256], *buf, *old = NULL;

  /* Server stable hash table variable's new value */
  rc = fscanf(file, "%s %s %d", name, entry, &value);
  check = rcvCheck(rc, 3, file, "stable hash table variable change");
  if (check != CHKOK) return(check); 
  c = fgetc(file);
  if (RoverDebug) fprintf(stderr, "[--] Entry size is %d\n", value);
  if (value) {
    if ((buf = (char *)malloc(value)) == NULL) {
      fprintf(stderr,"[--] Out of memory [checkLogStableHTVar:recover.c]\n");
      return(CHKERR);
    }
    if ((rc = fread(buf, sizeof(char), value, file)) != value) {
      fprintf(stderr,
	      "[--] Bad log stable hash table var read size %d [checkLogStableVar:recover.c]\n",
	      rc);
      free(buf);
      return(CHKERR);
    }
  } else {
    buf = malloc(1);
    buf[0] = '\0';
  }
  c = fgetc(file);
  if ((rc = stableHashSetEntry(name, entry, buf, value, &old)) != 0) {
    fprintf(stderr,
	    "[--] Got error %d during server stable hash table var change: %s[%s] %d:%s [checkLogStableVar:recover.c]\n",
	    rc, name, entry, value, buf);
    free(buf);
    return(CHKERR);
  }
  if (old) free(old);
  if (RoverDebug)
    fprintf(stderr,
	    "[--] Got server stable hash table log var change: %s[%s] %d:%s\n",
	    name, entry, value, buf);
  return(CHKOK);
}

PRIVATE CheckType checkLogStableHTDcl(FILE *file, ServerState **s)
{
  CheckType check;
  int created, rc;
  char name[256];

  /* Server stable hash table var declaration */
  rc = fscanf(file, "%s\n", name);
  check = rcvCheck(rc, 1, file, "stable hash table variable declaration");
  if (check != CHKOK) return(check); 
  created = stableHashInit(name);
  if (!created) {
    /* Found ht dcl, bad? */
    fprintf(stderr,
	    "[--] Ignoring DUPLICATE server stable log hash table var dcl: %s [checkLogStableHTDcl:recover.c]\n",
	    name);
    return(CHKERR);
  }
  if (RoverDebug) fprintf(stderr, "[--] Created hash table: %s\n", name);
  return(CHKOK);
}

PRIVATE CheckType checkLogData(FILE *file, ServerState **s)
{
  CheckType check;
  int i, rc, value, duplicate, psize, msgID, inc, seq, osize;
  char host[256], user[256], svc[256], cmd[256], ctype[256];
  char *buf, *pbuf, entry[256], c;
  int full = 0, empty = 0;

  /* Server logged message */
  rc = fscanf(file, "%d %s %s %d %d", &seq, host, user, &msgID, &inc);
  setSequenceNumber(seq);
  check = rcvCheck(rc, 5, file, "logged message [1]");
  if (check != CHKOK) return(check); 
  
  /* Find an empty slot and check for duplicates */
  for (duplicate = 0, *s = NULL, rc = -1, i = 0; i < MAXSTATES; i++) {
    if (States[i].inuse) {
      full++;
      if ((States[i].msgID == msgID) && (!strcmp(States[i].rhost, host)) &&
	  (!strcmp(States[i].ClientUser, user)) &&
	  (States[i].incarnation <= inc)) {
	/* Free earlier message of duplicate */
	duplicate = 1;
	reset_state(&(States[i]));
	if (rc == -1) rc = i;
      }
    } else {
      empty++;
      if (rc == -1) rc = i;
    }
  }
fprintf(stderr,"Full slots %d, empty slots %d, total slots %d\n",full,empty,i);
  i = rc;
  if (i == -1) {
    /* No slots, bad */
    fprintf(stderr, "[%d] Out of message slots [checkLogData:recover.c]\n", seq);
    return(CHKERR);
  }
  *s = &(States[i]);
  (*s)->first = (*s)->inuse = 1;
  (*s)->keepalive = (*s)->order = 0;
  (*s)->rhost = copyString(host);
  (*s)->ClientUser = copyString(user);
  (*s)->ServerSeq = seq;
  (*s)->msgID = msgID;
  (*s)->incarnation = inc;
  rc = fscanf(file, "%s %d %s %d %d %lu %d %d %s %d %s %d %d %d %d %d",
	      entry, &((*s)->port), svc, &((*s)->ServiceID), &((*s)->LogID),
	      &((*s)->UserTime), &((*s)->UserServer), &osize,cmd,&((*s)->size),
	      ctype, &((*s)->ContentLength), &((*s)->retry), &((*s)->compress),
	      &value, &psize);
  check = rcvCheck(rc, 16, file, "logged message [2]");
  if (check != CHKOK) return(check); 
  c = fgetc(file);
  (*s)->httpdHost = copyString(entry);
  (*s)->tcp = RoverTCPInit(-1, BUFSIZE);
  (*s)->Stdin = (*s)->Stdout = NULL;
  (*s)->Service = copyString(svc);
  (*s)->command = copyString(cmd);
  (*s)->ContentType = copyString(ctype);
  (*s)->reqType = ROVER_REQ;
  
  if (((*s)->Object = (char *)malloc(osize)) == NULL) {
    fprintf(stderr, "[%d] Out of memory [checkLogData:recover.c]\n", seq);
    return(CHKERR);
  }
  if ((rc = fread((*s)->Object, sizeof(char), osize, file)) != osize) {
    fprintf(stderr,
	    "[%d] Bad log msg read size %d [checkLogData:recover.c]\n",seq,rc);
    free((*s)->Object);
    return(CHKERR);
  }
  if (value) {
    if ((buf = (char *)malloc(value)) == NULL) {
      fprintf(stderr, "[%d] Out of memory [checkLogData:recover.c]\n", seq);
      return(CHKERR);
    }
    if ((rc = fread(buf, sizeof(char), value, file)) != value) {
      fprintf(stderr,
	      "[%d] Bad log msg read size %d [checkLogData:recover.c]\n",seq,rc);
      free(buf);
      return(CHKERR);
    }
  } else buf = NULL;
  (*s)->code = buf;
  if (psize) {
    if ((pbuf = (char *)malloc(psize)) == NULL) {
      fprintf(stderr, "[%d] Out of memory [checkLogData:recover.c]\n", seq);
      return(CHKERR);
    }
    if ((rc = fread(pbuf, sizeof(char), psize, file)) != psize) {
      fprintf(stderr, "[%d] Bad log msg read size %d [checkLogData:recover.c]\n",
	      rc, seq);
      free(pbuf);
      return(CHKERR);
    }
  } else pbuf = NULL;
  c = fgetc(file);
  (*s)->password = pbuf;

  (*s)->tcp = RoverTCPInit(-1, BUFSIZE);

  /* Check for messages that should be deleted (duplicates) */
  if (RoverDebug) {
    /* Save retry variable */
    i = (*s)->retry;
  }
  if (checkMsgID((*s)->ClientUser, (*s)->rhost, (*s)->msgID, (*s)->incarnation,
		 (*s)->retry) == -1) {
    duplicate = 1;
    reset_state(&(States[i]));
  }

  if (RoverDebug) {
    if (duplicate) {
      fprintf(stderr, "[%d] Got DUPLICATE server msg: %s@%s:%d (inc: %d, retry: %d)\n",
	      seq, user, host, msgID, inc, i);
    } else 
      fprintf(stderr,"[%d] Got server msg: %s@%s:%d (inc: %d)\n",seq, user, host, msgID,
	      inc);
  }
  return(CHKOK);
}

PRIVATE CheckType checkLogOutput(FILE *file, ServerState **s)
{
  CheckType check;
  int duplicate = 0, rc, total, seq, type, cache, size, lsize, dsize, csize;
  char c;

  /* Server logged output buffer */
  rc = fscanf(file, "%d", &seq);
  setSequenceNumber(seq);
  check = rcvCheck(rc, 1, file, "logged output buffer [1]");
  if (check != CHKOK) return(check); 
  if (!(*s = checkState(*s, seq))) {
    fprintf(stderr,
	    "[%d] Got LOG_OUTPUT before LOG_DATA [checkLogOutput:recover.c]\n",
	    seq);
    return(CHKERR);
  }

  rc = fscanf(file, "%d %d %d %d %d %d %d", &type, &csize, &dsize, &cache, &lsize, &size, &total);
  check = rcvCheck(rc, 7, file, "logged output buffer [2]");
  if (check != CHKOK) return(check); 
  c = fgetc(file);
  (*s)->ObjType = (RoverObjType) type;
  (*s)->loggedOutput = 1;
  (*s)->Cacheable = cache;
  (*s)->DVSize = size;
  (*s)->LogSize = lsize;
  (*s)->DataSize = dsize;
  (*s)->CodeSize = csize;
  (*s)->OutSize = total - state->DVSize;
  if ((*s)->output != NULL) {
    duplicate = 1;
    outstrReleaseBuffer((*s)->output);
    (*s)->output = NULL;
  }
  if ((*s)->OutSize) {
    outstrInit((*s)->OutSize);
    if ((rc = fread((*s)->output->buffer, sizeof(char), (*s)->OutSize, file))
	!= (*s)->OutSize) {
      fprintf(stderr,
	      "[%d] Bad log msg read size %d [checkLogOutput:recover.c]\n",seq,
	      rc);
      outstrReleaseBuffer((*s)->output);
      (*s)->output = NULL;
      return(CHKERR);
    }
  } else (*s)->output = NULL;
  if ((*s)->DVSize) {
    if ((rc = fread((*s)->DVbuf, sizeof(char), size, file)) != size) {
      fprintf(stderr,
	      "[%d] Bad log msg read size %d [checkLogOutput:recover.c]\n",seq,
	      rc);
      (*s)->DVbuf[0] = '\0';
      return(CHKERR);
    }
  } else (*s)->DVbuf[0] = '\0';
  c = fgetc(file);
  if (RoverDebug) {
    if (duplicate) {
      fprintf(stderr,
	      "[%d] Got DUPLICATE server log output buffer: %s@%s:%d (inc: %u, retry: %d)\n",
	      seq, (*s)->ClientUser, (*s)->rhost, (*s)->msgID,
	      (*s)->incarnation, (*s)->retry);
    } else 
      fprintf(stderr, "[%d] Got server log output buffer: %s@%s:%d (inc: %u)\n",
	      seq, (*s)->ClientUser, (*s)->rhost, (*s)->msgID, (*s)->incarnation);
  }
  return(CHKOK);
}

PRIVATE CheckType checkLogMsgID(FILE *file, ServerState **s)
{
  CheckType check;
  int rc, id, new, inc;
  char msgID[512];
  Tcl_HashEntry *hentry;
  ClientMsgID *cm;

  /* Server logged message identifier */
  rc = fscanf(file, "%s %d %d", msgID, &id, &inc);
  check = rcvCheck(rc, 3, file, "logged mssage identifier");
  if (check != CHKOK) return(check); 

  if ((cm = (ClientMsgID *)malloc(sizeof(ClientMsgID))) == NULL) return(CHKERR);
  cm->nextID = id;
  cm->incarnation = inc;
  hentry = Tcl_CreateHashEntry(&Info, msgID, &new);

  /* Update the expected msgID */
  Tcl_SetHashValue(hentry, (char *)cm);	/* Save it */

  if (RoverDebug) fprintf(stderr, "Got %smessage identifier: %s:%d (inc: %d)\n",
			  ((!new) ? "DUPLICATE " : ""), msgID, id, inc);
  return(CHKOK);
}

PUBLIC void recover(char *name, ServerState *s)
{
  CheckType check;
  int i, value, rc, valid = 1;
  char type[64];
  FILE *file;
  struct timeval RecoverStartTv, RecoverEndTv;

  /*
   * ROVER_LOG_VERSION <version#>
   * ROVER_LOG_START <serverSeq#>
   * ROVER_LOG_FORK <serverSeq#> <pid>
   * ROVER_LOG_RESULT <serverSeq#> <result#>
   * ROVER_LOG_RECOVER <serverSeq#> <recoveryProc>
   * ROVER_LOG_STABLE_VAR <serverSeq#> <name> <valueSize#>\n[data]\n
   * ROVER_LOG_STABLE_DCL <serverSeq#> <name> <valueSize#>\n[data]\n
   * ROVER_LOG_STABLE_HTDCL <name>
   * ROVER_LOG_STABLE_HTVAR <name> <entry> <valueSize#>\n[data]\n
   * ROVER_LOG_DATA <serverSeq#> <host> <user> <msgID#> <incarnation#>
   *		    <httpdHost> <port#> <service> <logID#> <userTime>
   *		    <server#> <osize> <command> <size> <contentType>
   *		    <contentLength> <retry> <compress> <csize>
   *		    <psize>\n[object][code][password]\n
   * ROVER_LOG_OUTPUT <serverSeq#> <ObjType#> <CSize#> <DSize#> <Cacheable#>
   *		      <LSize#> <DVSize#> <valueSize#>\n[data]\n
   * ROVER_LOG_MSGID <key> <msgID#> <incarnation#>
   */

  gettimeofday(&RecoverStartTv, &Tz);
  fprintf(stderr, "Reading recovery file %s...\n", name);
  ServerRecovering = READLOG;
  file = fopen(name, "r");
  while (valid) {
    if (feof(file)) {
      if (RoverDebug) fprintf(stderr, "End of log\n");
      break;
    }
    rc = fscanf(file, "%s\n", type);
    check = rcvCheck(rc, 1, file, "type");
    if (check == CHKEOF) break;
    else if (check == CHKERR) continue;

    if (RoverDebug) fprintf(stderr, "Type: %s\n", type);
    switch(getLOGtype(type)) {

    case ROVER_LOG_VERSION:
      if (checkLogVersion(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_START:
      if (checkLogStart(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_FORK:
      if (checkLogFork(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_RESULT:
      if (checkLogResult(file, &s) == CHKEOF) valid = 0;
      break;
      
    case ROVER_LOG_RECOVER:
      if (checkLogRecover(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_STABLE_VAR:
      if (checkLogStableVar(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_STABLE_DCL:
      if (checkLogStableDcl(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_STABLE_HTDCL:
      if (checkLogStableHTDcl(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_STABLE_HTVAR:
      if (checkLogStableHTVar(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_DATA:
      if (checkLogData(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_OUTPUT:
      if (checkLogOutput(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_MSGID:
      if (checkLogMsgID(file, &s) == CHKEOF) valid = 0;
      break;

    case ROVER_LOG_ERROR:
    default:
      fprintf(stderr, "Unknown log type (%s) [recover:recover.c]\n", type);
      break;

    }
  }
  fclose(file);
  
  gettimeofday(&RecoverEndTv, &Tz);
  fprintf(stderr, "Finished reading recovery file (time: %.4f)\n",
	  elapsedTime(RecoverStartTv, RecoverEndTv));
  /* Now check if there are any messages waiting */
  ServerRecovering = NORMALOP;
  for (state = NULL, value = i = 0; i < MAXSTATES; i++)
    if (States[i].inuse) {
      ServerRecovering = REDO;
      value++;
    }
  if (ServerRecovering != NORMALOP) 
    fprintf(stderr, "%d requests waiting to be rerun\n", value);
  else
    fprintf(stderr, "No requests need to be rerun\n");
}


/* Flush all user information and stable hash table information */
PRIVATE void flushTables()
{
  Tcl_HashSearch search, search2;
  Tcl_HashEntry *hentry, *hentry2;
  Tcl_HashTable *table;
  char *key;
  HashEntry *entry;
  UserData *data;

  /* Flush all user and stable hash table information */
  hentry = Tcl_FirstHashEntry(&Info, &search);
  while (hentry != NULL) {
    key = Tcl_GetHashKey(&Info, hentry);
    if (!strncmp(key, "MsgID_", 6)) {       /* Ignore message ID info */
      hentry = Tcl_NextHashEntry(&search);
      continue;
    } else if (!strncmp(key, "Hash_", 5)) { /* Delete stable hash table info */
      table = (Tcl_HashTable *)Tcl_GetHashValue(hentry);
      hentry2 = Tcl_FirstHashEntry(table, &search2);
      while (hentry2 != NULL) {
	entry = (HashEntry *)Tcl_GetHashValue(hentry2);
	free(entry);
	Tcl_DeleteHashEntry(hentry2);
	hentry2 = Tcl_FirstHashEntry(table, &search2);	/* Restart search */
      }
      Tcl_DeleteHashTable(table);
      Tcl_DeleteHashEntry(hentry);
      hentry = Tcl_FirstHashEntry(&Info, &search);	/* Restart search */
    } else if (!strncmp(key, "UserData_", 9)) {
      data = (UserData *)Tcl_GetHashValue(hentry);
      if (data->interp) Tcl_DeleteInterp(data->interp);
      Tcl_DeleteHashEntry(hentry);
      hentry = Tcl_FirstHashEntry(&Info, &search);	/* Restart search */
    }
  }
  globalTclInterp = NULL;
}


/* This procedure flushes all data from the log */
PRIVATE void flushLog()
{
  struct stat sbuf;

  fflush(ServerLogFile);
  fsync(ServerLogFD);
  fclose(ServerLogFile);

  /* Create the new log file */
  sprintf(command, "%s/rover-log-file", ServerLogDir);
  if (!stat(command, &sbuf)) unlink(command);	/* Delete the old file */
  if (roverInitLogFile(command, &ServerLogFile, &ServerLogFD)) {
    fprintf(stderr, "Open of file %s failed [flushLog:recover.c]\n",
	    command);
    exit(-1);
  }
  fflush(ServerLogFile);
  fsync(ServerLogFD);
  return;
}

/* 
 * Action format: {Sync/Async/None}__[Options__...]
 * 
 *		Async - Turns off ServerNoLog and ServerHardFlush
 *		Sync  - Turns off ServerNoLog and turns on ServerHardFlush
 *		None  - Turns on ServerNoLog and turns off ServerHardFlush
 * Options:
 *		NoHash       - Turns on ServerIgnoreHash and turns 
 *				off ServerUseHash
 *		IgnoreHash   - Turns on ServerUseHash and ServerIgnoreHash
 *		UseHash      - Turns on ServerUseHash and turns
 *				 off ServerIgnoreHash
 *
 *		NullTraceSV  - Turns on NullTrace and ServerUseStableVars
 *		NoStableVars - Turns off ServerUseStableVars
 *		StableVars   - Turn on ServerUseStableVars and turns
 *				 off NullTrace
 *
 *		Fail  	     - Turns on ServerDoFail
 *		NoFail       - Turns off ServerDoFail
 *
 *		LogDir__XX   - Sets ServerLogDir to next option
 *
 */
PUBLIC void LogChange(char *action)
{
  char *option, *next, *logMode = action;
  int doFlush = 0;

  option = index(logMode, '_');
  if (option) {
    *option = '\0';
    option += 2;
  }

  flushLog();
  if (!strcasecmp("Async", action)) {
    ServerNoLog = ServerHardFlush = 0;
  } else if (!strcasecmp("Sync", action)) {
    ServerNoLog = 0;
    ServerHardFlush = 1;
  } else if (!strcasecmp("None", action)) {
    ServerNoLog = 1;
    ServerHardFlush = 0;
  }

  while (option) {
    next = index(option, '_');
    if (next) {
      *next = '\0';
      next += 2;
    }
    if (!strcasecmp("NoHash", option)) {
      doFlush = 1;
      ServerIgnoreHash = 1;
      ServerUseHash = 0;
    } else if (!strcasecmp("IgnoreHash", option)) {
      doFlush = 1;
      ServerIgnoreHash = ServerUseHash = 1;
    } else if (!strcasecmp("UseHash", option)) {
      doFlush = 1;
      ServerIgnoreHash = 0;
      ServerUseHash = 1;
    } else if (!strcasecmp("NullTraceSV", option)) {
      doFlush = 1;
      NullTrace = ServerUseStableVars = 1;
    } else if (!strcasecmp("NoStableVars", option)) {
      doFlush = 1;
      ServerUseStableVars = 0;
    } else if (!strcasecmp("StableVars", option)) {
      doFlush = 1;
      NullTrace = 0;
      ServerUseStableVars = 1;
    } else if (!strcasecmp("Fail", option)) {
      ServerDoFail = 1;
    } else if (!strcasecmp("NoFail", option)) {
      ServerDoFail = 0;
    } else if (!strcasecmp("LogDir", option)) {
      if (!next) 
	fprintf(stderr, "Missing log directory [LogChange:recover.c]\n");
      else {
	ServerLogDir = copyString(next);
	next = index(next, '_');
	if (next) {
	  *next = '\0';
	  next += 2;
	}
      }
    } else 
      fprintf(stderr, "Unknown option `%s' [LogChange:recover.c]\n", option);
    option = next;
  }
  if (doFlush) flushTables();
  serverShowOptions();
}

PRIVATE time_t lastCleaned = 0;
PRIVATE int checkCleaned = 0;
#define MAXCLEANCHECK 20
#define MAXCLEANWAIT 120

/* Clean the server's log. Returns 1 the log was cleaned; otherwise,
 *  returns 1.
 */
PUBLIC int cleanLog()
{
  time_t now;
  int i, fd;
  FILE *file;
  Tcl_HashSearch search, search2;
  Tcl_HashEntry *hentry, *hentry2;
  Tcl_HashTable *table;
  char name[256], *key, *key2;
  HashEntry *entry;
  struct stat sbuf;

  if (logClean) return(0);
  if (++checkCleaned < MAXCLEANCHECK) return(0);
  checkCleaned = 0;
  time(&now);
  if (!lastCleaned) time(&lastCleaned);
  if ((now - lastCleaned) < MAXCLEANWAIT) return(0);

  fprintf(stderr, "Cleaning log at %s", ctime(&now));
  /* First, check for any pending messages. If there are any, abort */
  for (i = 0; i < MAXSTATES; i++) if (States[i].inuse) return(0);

  /* Next, close the log file */
  fflush(ServerLogFile);
  fsync(ServerLogFD);
  fclose(ServerLogFile);

  /* Create the new log file */
  sprintf(command, "%s/rover-log-file-NEW", ServerLogDir);
  if (!stat(command, &sbuf)) unlink(command);	/* Delete any old file */
  i = roverInitLogFile(command, &file, &fd);

  /* Dump all message ID and stable hash table information */
  for (hentry = Tcl_FirstHashEntry(&Info, &search); hentry != NULL;
       hentry = Tcl_NextHashEntry(&search)) {
    key = Tcl_GetHashKey(&Info, hentry);
    if (!strncmp(key, "MsgID_", 6)) {
      /* Save message ID info */
      ClientMsgID *cm = (ClientMsgID *) Tcl_GetHashValue(hentry);
      fprintf(file, "ROVER_LOG_MSGID %s %d %u\n", key, cm->nextID, cm->incarnation);
    } else if (!strncmp(key, "Hash_", 5)) {
      /* Save stable hash table info */

      fprintf(file, "ROVER_LOG_STABLE_HTDCL %s\n", key+5);
      table = (Tcl_HashTable *)Tcl_GetHashValue(hentry);
      for (hentry2 = Tcl_FirstHashEntry(table, &search2); hentry2 != NULL;
	   hentry2 = Tcl_NextHashEntry(&search2)) {
	entry = (HashEntry *)Tcl_GetHashValue(hentry2);
	key2 = Tcl_GetHashKey(table, hentry2);
	fprintf(file, "ROVER_LOG_STABLE_HTVAR %s %s %d\n", key+5, key2,
		entry->size);
	if (entry->size) {
	  if ((i = fwrite(entry->data, sizeof(char), entry->size, file)) != i){
	    fprintf(stderr,"Bad write size %d [cleanLog:recover.c]\n", i);
	    fclose(file);
	    unlink(command);
	    return(0);
	  }
	}
	fputc('\n', file);
      }
    } else if (!strncmp(key, "UserData_", 9)) {
      continue;
    } else {
      fprintf(stderr,
	      "Unknown key `%s' in Info hash table [cleanLog:recover.c]\n", 
	      key);
    } 
  }

  /* Create the new log file */
  fflush(file);
  fsync(fd);
  fclose(file);
  fprintf(stderr, "Closing NEW log file: %s\n", command);
  sprintf(name, "%s/rover-log-file", ServerLogDir);
  if (rename(command, name)) {
    perror("Rename failed:");
    unlink(command);
    return(0);
  }

  ServerLogFile = fopen(name, "a");
  if (ServerLogFile == NULL) {
    fprintf(stderr, "Open of log file %s failed [cleanLog:recover.c]\n", name);
    exit(-1);
  }
  ServerLogFD = fileno(ServerLogFile);
  time(&lastCleaned);
  logClean = 1;
  return(1);
}
