/**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: rmcache.c
 *
 * Description: Cache writer for ROVER browser interface layer
 *
 *
 * Global Functions:
 *    main
 *
 * Global Variables: none
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Alan deLespinasse and Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include "stdio.h"
#include "stdlib.h"
#include "errno.h"
#include "string.h"
#include "time.h"
#include "netdb.h"
#include "unistd.h"
#include "sys/types.h"
#include <sys/file.h>
#include <sys/fcntl.h>
#include <pwd.h>
#include "rover.h"
#include "base64.h"
#include "html.h"
#include "RoverTCP.h"

#define UNIQUELEN 15		/* maximum length of cache file names */

/* These are the same as in browse.c and rmgate.c. */
char *url=NULL,*pagetitle=NULL,*source=NULL,*sourcetitle=NULL,*path=NULL;
char *user=NULL,*hostname=NULL;
int form=0,visible=1,port=8080, prefetch = 0;
time_t time1=0,time2=0,time3=0;
int logID, compressLevel=0;
int image=0;
int intitle=0;			/* 1 if HTML title in progress */
size_t totalbytes=0;		/* of file read (without headers) */

/* Dummy definitions */
char *buffer;
int RoverDebug = 0;

/* Print everything we know to error log and exit. */
void puke(char* msg) {
  time_t t;
  char s[1000];

  t=time(NULL);
  fprintf(stderr,ctime(&t));
  fprintf(stderr,"Fatal error: %s\n",msg);
  fprintf(stderr,"Effective UID: %d\n",geteuid());
  fprintf(stderr,"Hostname: %s\n",hostname);
  fprintf(stderr,"Port: %d\n",port);
  fprintf(stderr,"URL: %s\n",url);
  fprintf(stderr,"Image: %s]n", image ? "yes" : "no");
  fprintf(stderr,"Log ID: %d\n", logID);
  fprintf(stderr,"Compress: %d\n", compressLevel);
  fprintf(stderr,"Pagetitle: %s\n",pagetitle);
  fprintf(stderr,"Source: %s\n",source);
  fprintf(stderr,"Sourcetitle: %s\n",sourcetitle);
  fprintf(stderr,"Path: %s\n",path);
  fprintf(stderr,"User: %s\n",user);
  fprintf(stderr,"Form: %s\n",form ? "yes" : "no");
  fprintf(stderr,"Visible: %s\n",visible ? "yes" : "no");
  fprintf(stderr,"Time1: %lu\n",(unsigned long) time1);
  fprintf(stderr,"Time2: %lu\n",(unsigned long) time2);
  fprintf(stderr,"Time3: %lu\n",(unsigned long) time3);
  fprintf(stderr,"Total bytes: %lu\n",(unsigned long) totalbytes);
  fprintf(stderr,"Remainder of stdin:\n");
  while (fgets(s,1000,stdin)) fputs(s,stderr);
  fprintf(stderr,"\n\n");
  fflush(stderr);
  exit(0);
}

/* Plain text callback for HTML parser. Sets the pagetitle global var if
   it's between a <title> and a </title>, as determined by the next
   function. */
void textcallback(struct HTMLparser* p) {
  char *t;

  if ((!intitle) || pagetitle) return;
  t=&(p->buf[strspn(p->buf," \t\r\n")]);
  if (t[0]==0) return;
  t=strtok(t,"\n\r");
  if (t==NULL || t[0]==0) return;
  pagetitle=strdup(t);
}

/* Tag callback for HTML parser. Keeps track of whether we're between
   a <title> and a </title>. */
void tagcallback(struct HTMLparser* p) {
  char *t=strtok(p->buf,"<> =\t\n\r\"");
  if (strcasecmp(t,"title")==0) intitle=1;
  else if (strcasecmp(t,"/title")==0) intitle=0;
}

int main() {
  char s[1000], buf[800], *line, *type=NULL;
  b64fromstate state;
  struct HTMLparser* parser = NULL;
  int linebeg, html=0, sock;
  size_t n;
  FILE *sockin,*sockout;
  
  time3=time(NULL);		/* keep records of timing data */
  
  freopen("rmcache_log","a",stderr);	/* redirect to error log */
  
/* parse SMTP headers */
  for (;;) {
    if (!fgets(s,1000,stdin)) puke("No header/body separator.");
    if (strspn(s," \t\n\r")==strlen(s)) break;
    
    if (strncasecmp(s,"x-url:",6)==0)
      url=strdup(strtok(&s[6]," \t\n\r"));
    else if (strncasecmp(s,"x-hostname: ",12)==0)
      hostname=strdup(strtok(&s[12]," \t\n\r"));
    else if (strncasecmp(s,"x-prefetch: ",11)==0)
      prefetch=atoi(strtok(&s[11]," \t\n\r"));
    else if (strncasecmp(s,"x-port: ",7)==0)
      port=atoi(strtok(&s[7]," \t\n\r"));
    else if (strncasecmp(s,"x-compress:",11)==0)
      compressLevel=atoi(strtok(&s[11]," \t\n\r"));
    else if (strncasecmp(s,"x-logid:",8)==0)
      logID=atoi(strtok(&s[8]," \t\n\r"));
    else if (strncasecmp(s,"x-pagetitle: ",13)==0)
      pagetitle=strdup(strtok(&s[13],"\n\r"));
    else if (strncasecmp(s,"x-source:",9)==0)
      source=strdup(strtok(&s[9]," \t\n\r"));
    else if (strncasecmp(s,"x-sourcetitle: ",15)==0)
      sourcetitle=strdup(strtok(&s[15],"\n\r"));
    else if (strncasecmp(s,"x-path:",7)==0)
      path=strdup(strtok(&s[7]," \t\n\r"));
    else if (strncasecmp(s,"x-form:",7)==0) {
      char* t=strtok(&s[7]," \t\n\r");
      form=(t[0]=='y' || t[0]=='Y') ? 1 : 0;
    } else if (strncasecmp(s,"x-image:",8)==0) {
      char* t=strtok(&s[8]," \t\n\r");
      image=(t[0]=='y' || t[0]=='Y') ? 1 : 0;
    } else if (strncasecmp(s,"x-user:",7)==0)
      user=strdup(strtok(&s[7]," \t\n\r"));
    else if (strncasecmp(s,"x-time1:",8)==0) {
      unsigned long ultime;
      sscanf(&s[8],"%lu",&ultime);
      time1=(time_t) ultime;
    } else if (strncasecmp(s,"x-time2:",8)==0) {
      unsigned long ultime;
      sscanf(&s[8],"%lu",&ultime);
      time2=(time_t) ultime;
    }
  }
  
  if (!url) puke("No url.");
  if (!user) puke("No user.");
  if (!hostname) puke("No hostname.");
  if (!port) port = 8080;

  /* Now open the connection to the browser interface application */
  sock = openTCP(hostname, port);
  if (sock == -1) puke("openTCP to browser failed.");

  sockin=fdopen(sock,"r");	/* Convert to streams. */
  sockout=fdopen(sock,"w");
  if (sockin==NULL || sockout==NULL) puke("fdopen(sock) failed.");

  fprintf(sockout,"CACHE %s HTTP/1.0\n",url);	/* update */

  if (!path) { /* means it wasn't explicitly requested - hide it */
    visible=0;
  }

  /* Send the header info. */
  sendHeader(sockout, url, compressLevel, image, logID, pagetitle, path,
	     time1, time2, time3, visible, prefetch);

  fprintf(stderr,"url: %s\n",url);
  fprintf(stderr,"image: %s\n",image ? "yes" : "no");


  /* Now send the data */
  b64fromreadinit(stdin,&state); /* decode it from base 64 */
  if (strncasecmp(url,"http:",5)==0) { /* only HTTP uses MIME headers */
    for (;;) {
      b64fromfgets(buf,1000,stdin,&state); /* process the MIME headers */
      totalbytes+=strlen(buf);
      fwrite(buf,1,strlen(buf),sockout);
      linebeg=strspn(buf," \t\r\n");
      if (linebeg==strlen(buf)) break;
      line=&buf[linebeg];
      if (strncasecmp(line,"content-type:",13)==0)
	type=strdup(strtok(&line[13]," \t\n\r"));
    }
  }
  /* This should also check for "application/html", etc. */
  if (type) {
    if ((strcasecmp(type,"text/html")==0) ||
	(strcasecmp(type,"application/html")==0)) 
      html=1;
  } else {
    if (strcasecmp(strrchr(url,'.'),".html")==0) 
      html=1;
  }
  if (html)
    parser=HTMLnewparse(textcallback,tagcallback,NULL);
  while ((n=b64fromread(buf,1000,stdin,&state)) != 0) { /* response body */
    fwrite(buf,1,n,sockout);
    if (html) HTMLparse(parser,buf,n);
    totalbytes+=n;
  }
  if (html) HTMLendparse(parser);
  b64fromend(&state);

  fprintf(stderr,"%lu %lu %lu %lu\n",time1,time2,time3,
	  (unsigned long)totalbytes);
  fflush(stderr);

  fflush(sockout);
#if 0
  {
    int respcode;
    fgets(s,1000,sockin);
    sscanf(s,"HTTP/1.0 %d",&respcode);
    if (respcode != 200)
      puke("Bad response code");
  }
#endif  
  fclose(sockin);
  fclose(sockout);
  close(sock);

/* Save stats on transfer times, etc. */
#ifdef KEEPLOG
  sockout=fopen(".log","a");
  if (sockout) {
    fprintf(sockout,"%lu %lu %lu %lu\n",time1,time2,time3,totalbytes);
    fclose(sockout);
  }
#endif

  return 0;
}
