 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: rover.c
 *
 * Description: Main Rover Server module.
 *		This program is intended to be setuid root. It operates in
 *		one of two modes: as a forked, cgi-bin application and as
 *		a standalone server. When operating as a forked server, it
 *		accepts	a cgi-bin-style QUERY_STRING in the environment and
 *		parses it for several query elements. The elements include
 *		the USER and AUTH elements, which are used to perform an
 *		authentication validation check of the user. If successful,
 *		the server will configure the user's gid, uid, and
 *		environment and attempt to execute the desired operation.
 *
 *
 * Global Functions: See HTML documentation
 * Global Variables: See HTML documentation
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

/* #define AUTHDEBUG */
/* #define EXEC */
/* #define TIMING */
/* #define TIMER */

#include "rover.h"
#include "rovertcp.h"
#include "database.h"
#include "RoverZlib.h"
#include "RoverEscape.h"
#include <grp.h>
#include <sys/types.h>
#include "RoverACL.h"
#include "RoverConfig.h"

#ifndef LF
#define LF 10
#endif /* LF */
#ifndef CR
#define CR 13
#endif /* CR */

double TotalServerTime = 0.0, TotalAppTime = 0.0;
int TotalTimeCount = 0;

FILE *ServerLogFile = NULL;
FILE *ServerLogFile2 = NULL;
int ServerNoLog = 0;
int ServerHardFlush = 0;
int ServerLogFD = -1;
int ServerLogFD2 = -1;
char *ServerLogDir = NULL;
ServerState States[MAXSTATES];
ServerState *TCPStates[MAXSTATES];
Tcl_HashTable Info;
Tcl_HashEntry *hentry;
int RoverDebug = 0;
ServerRecoverType ServerRecovering = NORMALOP;
struct timeval  AppStartTv, AppEndTv;
struct timeval  ServerStartTv, ServerEndTv;

int pending = 0;
int NoAuth = 0;
char *ServerName = NULL;
int standalone = 0;
uid_t RoverUid = -1;
gid_t RoverGid = -1;
char *RoverHomeEnv = NULL;
char *RoverUserEnv = NULL;
int RoverRootID = -1;

PRIVATE char line[8194], command[256], output[2048], authtmp[256];
extern int statsAdd(double value);
PRIVATE void RoverSendOutput(ServerState *s);

PUBLIC void RoverInitRequest(ServerState *s)
{
  state = s;

  roverResetAuth(s);
  s->DVSize = s->CommandDone = 0;
  s->status = 200;
  s->Cacheable = ROVER_BEST_EFFORT;

  if (!s->ClientUser) {
    s->status = 401;
    errorExit("No username specified");
  }

  if (!s->rhost) {
    s->status = 401;
    errorExit("No hostname specified");
  }

  if (!s->httpdHost) {
    s->status = 401;
    errorExit("No httpd hostname specified");
  }

  if (!s->Service) {
    s->status = 401;
    errorExit("No service specified");
  }

  perform_auth_check(s);
  outstrInit(INITSIZE);

  if (s->LogID == -1) {
    fprintf(stderr, "No logid found!\n");
    s->LogID = 0;
  }

  if (!s->Object) {
    s->status = 400;
    errorExit("No object (OBJ) string found");
  }
  if (RoverDebug) fprintf(stderr, "Got dv: <%u:%u> for `%s'\n", s->UserServer, 
			  (unsigned int) s->UserTime, s->Object);
  if (ServerID != s->UserServer) {
    s->UserTime = (time_t) 0;
    s->UserServer = 0;
  }
}

#if defined(TIMER) || defined(TIMEEXEC)
PRIVATE void RoverServerEndTime()
{
#ifdef TIMER
  gettimeofday(&ServerEndTv, &Tz);
  TotalServerTime += elapsedTime(ServerStartTv, ServerEndTv);
  fprintf(stderr, "Time at server: %.4f\n", elapsedTime(ServerStartTv,
							ServerEndTv));
  TotalTimeCount++;
#endif /* TIMER */
}
#endif /* TIMER || TIMEEXEC */

PUBLIC void rovermain(ServerState *s, int argc, char **argv)
{
  state = s;
  RoverInitRequest(s);
  RoverSetAuth(s);

  if (standalone && (ServerRecovering != NORMALOP)) {
    /* Run the recovery procedure */
    if (s->auth.Nobody)
      /* No state is changed by this command; so just rerun it */
      perform_nobody_command(s);
    else
      recover_command(s);
  } else {
    /* Normal invocation */
    if (s->auth.Nobody)
      perform_nobody_command(s);
    else
      perform_command(s);
  }
  
  if (s->CommandDone) {		/* Return if nothing left to do */
    TotalAppTime += elapsedTime(AppStartTv, AppEndTv);
    fprintf(stderr, "Time in application: %.4f\n", elapsedTime(AppStartTv,
							       AppEndTv));
#ifdef TIMER
    RoverServerEndTime();
#endif TIMER
    return;
  }

  RoverSendOutput(s);
}

PRIVATE void RoverSendOutput(ServerState *s)
{
  char buf[128];
  int rc, extra;

  if (s->EmittingData) {
    /* Allocate a placeholder for the data null */
    outstrD("X");
  } else if (s->EmittingLog) { /* Update log record count information */
    sprintf(buf, "%7d", s->LogCount);
    memcpy(s->LogStart, buf, 7);
  }

   /* Null terminate buffer */
  s->output->buffer[s->DataSize+s->CodeSize+s->LogSize] = '\0';

#ifdef TIMEEXEC
    RoverServerEndTime();
#endif TIMEEXEC

  if (RoverDebug)
    fprintf(stderr, "Sending %d code, %d data, and %d log bytes (%d actual)\n",
	    s->CodeSize, s->DataSize, s->LogSize,
	    (int) (strlen(s->output->buffer)+1));

  sprintf(s->DVbuf, "%u_%u", ServerID, (unsigned int) s->dvtime);
  s->DVSize = strlen(s->DVbuf)+1;	/* Null terminated */

  rc = sendServerResponse(&extra, 0);
  if (RoverDebug) fprintf(stderr, "Actually wrote %d+%d bytes\n", rc, extra);
#ifdef TIMER
  TotalAppTime += elapsedTime(AppStartTv, AppEndTv);
  fprintf(stderr, "Time in application: %.4f\n", elapsedTime(AppStartTv,
							     AppEndTv));
  RoverServerEndTime();
#endif /* TIMER */
}

PUBLIC double elapsedTime(struct timeval tv1, struct timeval tv2) 
{
  double start = ((double) tv1.tv_sec) * 1000000.0 + tv1.tv_usec;
  double stop = ((double) tv2.tv_sec) * 1000000.0 + tv2.tv_usec;
  double seconds = (stop - start) / 1000000.0;

  return(seconds);
}

PUBLIC void perform_nobody_command(ServerState *s)
{
  CMDtype cmd;
  char *value, *ptr, *name, *class, *type;
  struct stat sbuf;
  struct passwd *entry;
  FILE *file;
  int i, rc, extra, size, n;
  z_stream *c_stream = NULL;	/* Compression stream */

  /* Do the command */

  if (!s->command) {
    s->status = 400;
    errorExit("No command (CMD) string found");
  }
  cmd = Name2Cmd(s->command);

  switch(cmd) {
  case TEST:
    if (!s->size) {
      s->status = 400;
      errorExit("No size (SIZE) string found");
    }
    testExit(atoi(s->Object));
    return;
    break;

  case HTTP:
    gettimeofday(&AppStartTv, &Tz);
    processHTTP(s, 0, NULL, 0);
    gettimeofday(&AppEndTv, &Tz);
    fprintf(stderr, "HTTP time is %f\n", elapsedTime(AppStartTv, AppEndTv));
    statsAdd(elapsedTime(AppStartTv, AppEndTv));
    break;

  case RESET:
    errorExit("Can't perform RESET as nobody [perform_nobody_command:rover.c]\n");
    break;

  case IMPORTBINARY:
    s->ObjType = ROVER_BINARY;
    if (strstr(s->Object, "..")) {
      s->status = 404;
      errorExit("Object name contains illegal `..' sequence");
    }
    if (*(s->Object) != '~') {
      strcpy(command, RoverRoot[RoverRootID]);
      strcat(command, s->Object);
    } else {
      name = ++(s->Object);
      if ((ptr = index(s->Object, '/')) == NULL) {
	s->status = 404;
	errorExit("Illegal username in URL");
      }
      *ptr = '\0';
      if ((entry = getpwnam(name)) == NULL) {
	s->status = 401;
	errorExit("Invalid username in URL");
      }
      sprintf(command, "%s/public_html/%s", entry->pw_dir, ptr+1);
    }

    if (!stat(command, &sbuf)) {
      if ((sbuf.st_mode & (S_IWUSR|S_IWGRP|S_IWOTH)) == 0) {
	fprintf(stderr, "Read-only object ==> ROVER_CACHE_FOREVER\n");
	s->Cacheable = ROVER_CACHE_FOREVER;
	s->UserTime = (time_t) 0;
      }
      s->dvtime = sbuf.st_mtime;
      if (checkVerify(command, s, ROVER_BINARY_UPDATE)) return;
    } else {
      sprintf(line, "Object (%s) not found", command);
      s->status = 404;
      errorExit(line);
    }
    
    s->CodeSize = sbuf.st_size;

    fprintf(stderr, "Sending %lu binary code bytes\n",
			    (unsigned long) sbuf.st_size);
    sprintf(s->DVbuf, "%u_%u", ServerID, (unsigned int) s->dvtime);
    s->DVSize = strlen(s->DVbuf)+1;	/* Null terminated */

    rc = sendServerResponse(&extra, 1);
    rc = 0;

    if (s->compress) {
      if (!(c_stream = InitCompress(s->compress))) {
	s->status = 500;
	errorExit("InitCompress error");
      }
    }
    if ((file = fopen(command,"r")) == NULL) {
      sprintf(line, "File open failed: %s\n", strerror(errno));
      s->status = 404;
      errorExit(line);
    }
    while (1) {
      i = fread(line, sizeof(char), 8192, file);
      if (s->compress) {
	n = 0;
	if (Compress(c_stream, s->Stdout, line, i, output, 8192, &n)
	    != Z_OK) {
	  s->status = 500;
	  errorExit("Compress error");
	}
	rc += n;
      } else {
	fwrite(line, sizeof(char), i, s->Stdout);
	rc += i;
      }
      if (feof(file)) break;
    }
    if (s->compress) {
      if (EndCompress(c_stream, s->Stdout, output, 8192, &n) != Z_OK) {
	s->status = 500;
	errorExit("EndCompress error");
      }
      rc += n;
    }
    fclose(file);
    fprintf(stderr, "Binary send complete\nActually wrote %d+%d bytes\n", rc,
	    extra);
    if (!s->compress && (rc != sbuf.st_size)) {
      fprintf(stderr, "Expected to send %lu, but sent %d instead\n",
	      sbuf.st_size, rc);
      abort();
    }
    s->CommandDone = 1;
    return;
    break;

  case IMPORTDATA:
    /* Ordinary object */
    s->ObjType = ROVER_TCL_TK;
    if (strstr(s->Object, "..")) {
      s->status = 404;
      errorExit("Object name contains illegal `..' sequence");
    }
    
    if(!strncmp("Rover__tclIndex__", s->Object,
		strlen("Rover__tclIndex__"))) {
      /* Request for Rover__tclIndex object */
      tcl_importdata(s->Object);
    } else {
      strcpy(command, RoverRoot[RoverRootID]);
      strcat(command, "/");
      strcat(command, s->Object);
      
      if (!stat(command, &sbuf)) {
	if ((sbuf.st_mode & (S_IWUSR|S_IWGRP|S_IWOTH)) == 0) {
	  fprintf(stderr, "Read-only object ==> ROVER_CACHE_FOREVER\n");
	  s->Cacheable = ROVER_CACHE_FOREVER;
	  s->UserTime = (time_t) 0;
	}
	s->dvtime = sbuf.st_mtime;
	if (checkVerify(command, s, ROVER_TCL_TK_UPDATE)) return;
	outstrResize(sbuf.st_size+10);
      } else {
	sprintf(line, "No object found (%s) [rover.c]!", command);
	s->status = 404;
	errorExit(line);
      }
      
      file = fopen(command,"r");
      if (file == NULL) {
	sprintf(line, "File open failed: %s\n", strerror(errno));
	s->status = 404;
	errorExit(line);
      }
      emitData();
      i = 0;
      while (1) {
	line[0] = '\0';
	i += (rc = fread(line, sizeof(char), 1023, file));
	line[rc] = '\0';
	outstrD(line);
	if (feof(file)) break;
      }
      fclose(file);
      if (i != sbuf.st_size) {
	fprintf(stderr, "Only read %d, instead of %lu\n", i, sbuf.st_size);
	abort();
      }
    }
    break;

  case IMPORTCMD:
    /* Check if the object is a null qrpc */
    if (!strncmp("QRPC", s->Object, strlen("QRPC"))) {
      s->ObjType = ROVER_TCL_TK;
      s->Cacheable = ROVER_NO_CACHE;
 
      /* Check if the object is a timer object */
    } else if (!strncmp("Time__", s->Object, strlen("Time__"))) {
      value = copyString(s->Object);
      s->ObjType = ROVER_TCL_TK;
      type = index(value, '_');
      type += 2;
      if (!strncmp("NC__", type, strlen("NC__"))) {
	s->Cacheable = ROVER_NO_CACHE;
      } else {
	s->Cacheable = ROVER_CACHE_FOREVER;
      }
      class = index(type, '_');
      *class = '\0';
      class += 2;
      size = atoi(class);
      outstrVC("proc Time__%0__%1 {} {}\nreturn \"nullCallback\"\n", 2,
	       type, class);
/*      fprintf(stderr, "Making timing object w/ %d code and %d data bytes\n",
	      s->CodeSize+1, size); */
      /*    size -= (s->CodeSize+2);
      if (size < 0) return; */
      /* Now we make an object */
      emitData();
      /* Adjust the buffer's size */
      outstrResize(size);
      while (size--)
	outstrD("X");
      free(value);

      /* Check if the object is a Rover object */
    } else if (strncmp("Rover__", s->Object, strlen("Rover__"))) {
      /* Nope, must be an ordinary object */
      s->ObjType = ROVER_TCL_TK;

      strcpy(command, RoverRoot[RoverRootID]);
      strcat(command, "/");
      strcat(command, s->Object);

      if (!stat(command, &sbuf)) {
	if ((sbuf.st_mode & (S_IWUSR|S_IWGRP|S_IWOTH)) == 0) {
	  fprintf(stderr, "Read-only object ==> ROVER_CACHE_FOREVER\n");
	  s->Cacheable = ROVER_CACHE_FOREVER;
	  s->UserTime = (time_t) 0;
	}
	s->dvtime = sbuf.st_mtime;
	if (checkVerify(command, s, ROVER_TCL_TK_UPDATE)) return;
	fprintf(stderr,"Object `%s' size is %lu\n",s->Object,sbuf.st_size);
	outstrResize(sbuf.st_size+10);
      } else {
	sprintf(line, "No object found (%s)!", command);
	s->status = 404;
	errorExit(line);
      }

      if ((file = fopen(command,"r")) == NULL) {
	sprintf(line, "File open failed: %s\n", strerror(errno));
	s->status = 404;
	errorExit(line);
      }
      i = 0;
      while (1) {
	line[0] = '\0';
	i += (rc = fread(line, sizeof(char), 1023, file));
	line[rc] = '\0';
	outstrC(line);
	if (feof(file)) break;
      }
      fclose(file);
      if (i != sbuf.st_size) {
	fprintf(stderr, "Only read %d, instead of %lu\n", i, sbuf.st_size);
	abort();
      }
    } else {
      sprintf(line, "User nobody can't import Rover class (%s) objects\n",
	      s->Object);
      errorExit(line);
    }
    break;

  case EXPORTCMD:
    fprintf(stderr, "Got export operation `%s'\n", s->Object);
    /* Check if the object is a null qrpc */
    if (!strncmp("QRPC", s->Object, strlen("QRPC"))) {
      s->ObjType = ROVER_TCL_TK;
      s->Cacheable = ROVER_NO_CACHE;
 
      /* Check if the object is a timer object */
    } else if (!strncmp("Time__", s->Object, strlen("Time__"))) {
      value = copyString(s->Object);
      s->ObjType = ROVER_TCL_TK;
      type = index(value, '_');
      type += 2;
      s->Cacheable = ROVER_NO_CACHE;
      class = index(type, '_');
      *class = '\0';
      class += 2;
      outstrVC("proc Time__F__%1 {} {}\nreturn \"nullCallback\"\n", 1, class);
      fprintf(stderr, "Timing object export: %d data bytes\n",
	      s->CodeSize+1);
      free(value);

      /* Check if the object is a Rover object */
    } else if (strncmp("Rover__", s->Object, strlen("Rover__"))) {
      /* Nope, must be an ordinary object */
      errorExit("Can't export ordinary obj [perform_nobody_command:rover.c]\n");
    }
    break;
  default:
    errorExit("Invalid command for user nobody\n");
    break;
  }
  fflush(s->Stdout);
}

PUBLIC void initState(ServerState *s)
{
  s->inuse = 1;
  s->incarnation = -1;
  s->msgID = -1;
  s->ClientUser = NULL;
  s->rhost = NULL;
  s->Service = NULL;
  s->ServiceID = 0;
  s->LogID = -1;
  s->UserTime = (time_t) 0;
  s->UserServer = 0;
  s->Object = NULL;
  s->command = NULL;
  s->password = NULL;
  s->code = NULL;
  s->size = 0;
  s->Query = NULL;
  s->QuerySize = 0;
  s->retry = 0;
  s->pid = 0;
  s->httpdHost = NULL;
  s->port = 0;
  s->recoverProc = NULL;
  s->openConnection = 0;
  s->first = 0;
  s->keepalive = 0;
  s->loggedOutput = 0;
  s->outputBuffer = NULL;
  s->output = NULL;
  s->Header[0] = '\0';
  s->batch = 0;
  s->count = 0;
  s->bufferOffset = 0;
  s->remoteName = NULL;
  s->remoteIP = NULL;
  s->inflate = NULL;
  s->plugInData = NULL;

  s->Home = s->FastHome = NULL;
  s->loggedOutput = 0;
}

PUBLIC void print_query(ServerState *s)
{
  fprintf(stderr, "Query information:\nIncarnation: %d\n", s->incarnation);
  fprintf(stderr, "MsgID: %d\n", s->msgID);
  fprintf(stderr, "ClientUser: %s\n", s->ClientUser);
  fprintf(stderr, "rhost: %s\n", s->rhost);
  fprintf(stderr, "httpHost/port: %s/%d\n", s->httpdHost, s->port);
  fprintf(stderr, "Service: %s (%d)\n", s->Service, s->ServiceID);
  fprintf(stderr, "LogID: %d\n", s->LogID);
  fprintf(stderr, "UserTime: %lu\n", (unsigned long) s->UserTime);
  fprintf(stderr, "UserServer: %d\n", s->UserServer);
  fprintf(stderr, "Object: %s\n", s->Object);
  fprintf(stderr, "command: %s\n", s->command);
  fprintf(stderr, "password: %s\n", s->password);
  fprintf(stderr, "size: %d\n", s->size);
  fprintf(stderr, "retry: %d\n", s->retry);
  fprintf(stderr, "code: %s\n", s->code);
}

PUBLIC void process_query(ServerState *s)
{
  char *q = NULL, *name, *value;
  ServerState *first = s;
  int i, save = 0;

  if (!(q = s->Query)) {
    first->status = 400;
    errorExit("No query information to decode");
  }

 if (RoverDebug) fprintf(stderr, "Query string is |%s|\n", q);
  
  while (*q) {
    if (*q == '#') {			/* Found batched query separator */
      s->keepalive = 1;			/* Make sure previous state knows */
      if (s->rhost && !s->httpdHost) s->httpdHost = s->rhost;
      if (!s->code) s->code = "";
      if (standalone) {
	save = 0;
	saveState(s);
      }
      if (RoverDebug) print_query(s);
      first->batch++;
      q++;
      for (i = 0; i < MAXSTATES; i++)
	if (!States[i].inuse) break;
      if (i == MAXSTATES) {
	/* We should have a more graceful choice here */
	first->status = 500;
	errorExit("Out of message slots\n");
      }
      s = &(States[i]);
      initState(s);
      s->compress = first->compress;
      s->tcp = first->tcp;
      s->callback = first->callback;
      s->reqType = first->reqType;
      s->Stdin = first->Stdin;
      s->Stdout = first->Stdout;
      s->Query = first->Query;		/* Last one will free it */
    }
    save = 1;
    name = q;				/* We start with the name */
    while (*q && (*q != '='))		/* Look for separator */
      q++;
    if (!(*q)) {			/* Check that we didn't run off end */
      fprintf(stderr, "WARNING! Incomplete request name %s\n", name);
      break;
    } else {
      *q = '\0';			/* Null terminate the name */
      RoverUnescape(name);		/* Convert it */
      s->QuerySize++;			/* Another entry */
      
      value = ++q;
      while (*q && (*q != '&'))		/* Look for separator */
	q++;
      if (*q) {				/* See if we reached the query's end */
	*q = '\0';			/* Null terminate the name */
	q++;
      }
      RoverUnescape(value);		/* Convert it */

      if (!strcmp(name, ROVER_MSGID_SPEC)) {
	s->msgID = atoi(value);
      } else if (!strcmp(name, ROVER_INC_SPEC)) {
	s->incarnation = atoi(value);
      } else if (!strcmp(name, ROVER_CMD_SPEC)) {
	s->command = value;
      } else if (!strcmp(name, ROVER_AUTH_SPEC)) {
	s->password = value;
      } else if (!strcmp(name, ROVER_USER_SPEC)) {
	s->ClientUser = value;
      } else if (!strcmp(name, ROVER_HOST_SPEC)) {
	s->rhost = value;
      } else if (!strcmp(name, ROVER_SVC_SPEC)) {
	s->Service = value;
      } else if (!strcmp(name, ROVER_SVCID_SPEC)) {
	s->ServiceID = atoi(value);
      } else if (!strcmp(name, ROVER_LOGID_SPEC)) {
	s->LogID = atoi(value);
      } else if (!strcmp(name, ROVER_TIME_SPEC)) {
	sscanf(value,"%u_%u",&(s->UserServer),(unsigned int *) &(s->UserTime));
      } else if (!strcmp(name, ROVER_OBJ_SPEC)) {
	s->Object = value;
      } else if (!strcmp(name, ROVER_CODE_SPEC)) {
	s->code = value;
      } else if (!strcmp(name, ROVER_SIZE_SPEC)) {
	s->size = atoi(value);
      } else if (!strcmp(name, ROVER_RETRY_SPEC)) {
	s->retry = atoi(value);
      } else if (!strcmp(name, ROVER_HTTPD_SPEC)) {
	s->httpdHost = value;
      } else if (!strcmp(name, ROVER_PORT_SPEC)) {
	s->port = atoi(value);
      } else {
	fprintf(stderr, "WARNING! Ignoring unknown query element: %s", name);
      }
    }
  }
  if (s->rhost && !s->httpdHost) s->httpdHost = s->rhost;
  if (!s->code) s->code = "";
		 
  if (first->batch) pending = 1;
  if (RoverDebug) {
    fprintf(stderr, "Query batch size: %d\n", first->batch+1);
    print_query(s);
  }
  if (standalone && save) saveState(s);
}

PUBLIC void perform_auth_check(ServerState *s)
{
  char *buf, *authName = NULL;
  struct passwd *entry = NULL;
  int idx;

  if (!strncasecmp(s->ClientUser, "nobody", strlen("nobody"))) {
    s->auth.Nobody = 1;
    authName = "nobody";
  }

  if (!s->password) {
    s->status = 401;
    errorExit("AUTH: No authorization specified");
  }

  if (!strcasecmp(s->ClientUser, "root")) {
    s->status = 403;
    errorExit("AUTH: Root use not allowed");
  }

  if (!s->auth.Nobody) {	/* Check for valid user */
    authName = s->ClientUser;
    for (idx = 0, buf = RoverClientACL[0];
	 buf && strcmp(s->ClientUser, buf);
	 buf = RoverClientACL[++idx]);
    if (strlen(buf) == 0){
      s->status = 403;
      errorExit("AUTH: Invalid authorization");
    }
  }

#ifdef AUTHDEBUG
  fprintf(stderr, "AUTH: Set auth to %s\n", authName);
#endif /* AUTHDEBUG */

  set_serverID();
  idx = 0;
 retry:
  if ((entry = getpwnam(authName)) == NULL) {
      if (idx < 2) {
	  fprintf(stderr, "AUTH: Tried name `%s'\n", authName);
	  idx++;
	  setpwent();
	  goto retry;
	  }
      s->status = 401;
      errorExit("AUTH: Invalid authorization");
  }
  

#ifdef DOAUTH
  if (strcmp(crypt(s->password, entry->pw_passwd), entry->pw_passwd)) {
    s->status = 401;
    errorExit("AUTH: Invalid authorization [getpw]");
  }
#endif /* DOAUTH */


  s->auth.gid = entry->pw_gid;
  s->auth.uid = entry->pw_uid;

  /* Set user-specific information */
  s->FastHome = (char *)malloc(5+strlen(entry->pw_name)+1);
  if (!s->FastHome) fatalError("AUTH: Server unable to allocate fast home variable");
  sprintf(s->FastHome, "/u1/%s", entry->pw_name);

  s->Home = copyString(entry->pw_dir);
  s->auth.HomeEnv = (char *)malloc(strlen("HOME=")+strlen(entry->pw_dir)+1);
  if (!s->auth.HomeEnv) 
    fatalError("AUTH: Server unable to allocate environment var. buffer");
  sprintf(s->auth.HomeEnv, "HOME=%s", entry->pw_dir);

  s->auth.UserEnv=(char *)malloc(strlen("USER=")+strlen(entry->pw_name)+1);
  if (!s->auth.UserEnv)
    fatalError("AUTH: Server unable to allocate environment var. buffer");
  sprintf(s->auth.UserEnv, "USER=%s", entry->pw_name);
}

PUBLIC void RoverSetAuth(ServerState *s)
{
  UserData *data;

  /* Switch to the appropriate groups and uid */
  if (!NoAuth) {
    if (initgroups((s->auth.Nobody ? "nobody" : s->ClientUser), s->auth.gid)) {
      sprintf(line,
	      "AUTH: Server initgroups failed: %s. Check that the server is running as root",
	      strerror(errno));	
      fatalError(line);
    }
  }
#ifdef AUTHDEBUG
  fprintf(stderr, "AUTH: uid is %d\n", s->auth.uid);
#endif /* AUTHDEBUG */

  if (!NoAuth) {
#ifdef linux
      if (s->auth.uid != 65535) {   /* Linux doesn't like -1 */
#endif /* linux */
	if (seteuid(s->auth.uid)) {	
	  sprintf(line,
		  "AUTH: Server seteuid failed: %s Check that the server is running as root",
		  strerror(errno));	
	  fatalError(line);
	}
#ifdef linux
      }
#endif /* linux */
  }
  
  /* Set user-specific information */
  putenv(s->auth.HomeEnv);
  putenv(s->auth.UserEnv);
  if (s->auth.Nobody) return;
  
  if (standalone) {
    sprintf(authtmp, "UserData_%s@%s_%s", s->ClientUser, s->rhost, s->Service);
    hentry = Tcl_FindHashEntry(&Info, authtmp);
    if (hentry && (data =(UserData *)Tcl_GetHashValue(hentry)) && data->interp)
      globalTclInterp = data->interp;
  }
}

PUBLIC void reset_auth()
{
  /* Switch to the appropriate groups and uid */
  if (setreuid(RoverUid, RoverUid)) 
    fprintf(stderr,
	    "AUTH: Unrecoverable fatal error: setreuid for root failed!");

  if (!NoAuth) {
    if (initgroups("root", RoverGid))
      fprintf(stderr,
	      "AUTH: Unrecoverable fatal error: initgroups for root failed!");
  }
#ifdef AUTHDEBUG
  fprintf(stderr, "AUTH: Reset auth to root\n");
#endif /* AUTHDEBUG */
}

PUBLIC void reset_arp(ServerState *s, char *host)
{
  struct stat sbuf;
  char *value;
  char output[512];

  value = RoverEscapeShellCmd(host);
  fprintf(stderr, "Flushing arp cache entry for %s...\n", value);
  sprintf(output, "/etc/arp");
  if (stat(output, &sbuf)) {
    sprintf(output, "/sbin/arp");
    if (stat(output, &sbuf)) {
      fprintf(stderr, "arp command not found!");
      return;
    }
  }
  strcat(output, " -d ");
  strcat(output, value);
  reset_auth();		/* Bad, become root so we can perform arp command :( */
  fprintf(stderr, "Command is `%s'\n", output);
  system(output);
  perform_auth_check(s);	/* Take us back, quickly! */
  RoverSetAuth(s);
  if (value != host) free(value);
  fprintf(stderr, "flush complete\n");
}

/*
 * Generate a response status header.
 */
PUBLIC void generateServerHeader()
{
  state->Header[0] = '\0';
  if (!state->openConnection) {
    if (state->first) {
#ifdef NPH
      sprintf(state->Header, "HTTP/1.0 %d %s%c%c%s%c%c", state->status,
	      StatusMsg(state->status), CR, LF,
	      (standalone ? "Server: Standalone Rover/$Revision: 1.118 $" :
	       "Server: NCSA/1.4.1+Rover/$Revision: 1.118 $"), CR, LF);
#endif /* NPH */
      sprintf(line, "Content-type: application/x-object%c%c%c%c", CR, LF, CR, LF);
      strcat(state->Header, line);
    }
  } else {
    /*
     * Generate a header for the client's httpd
     *		ROVER <url> <session> <sessionID> [<logID>]
     */
    sprintf(state->Header, "ROVER %s %s %d %d\n", state->Object,
	    state->Service, state->ServiceID, state->LogID);
  }

  sprintf(line, "%2d %6d %7d %9d %1d %1d %1d %7d %4d\n",
	  state->ObjType, state->LogID, state->CodeSize, state->DataSize,
	  state->keepalive, state->compress, state->Cacheable, state->LogSize,
	  state->DVSize);
  strcat(state->Header, line);
}

void RoverRootDirsInit()
{
  struct stat sbuf;

  /* Check for rover directory */
  if (RoverRootID < 0) {
    for(RoverRootID = 0; RoverRoot[RoverRootID] != NULL; RoverRootID++)
      if (stat(RoverRoot[RoverRootID], &sbuf) == 0) break;
    if(RoverRootID < 0) {
      fprintf(stderr, "RoverRoot directory not found");
      exit(-1);
    }
  }
}

#if 0
/*
 * Construct and send a status header to the client, followed
 *   by the output buffer and any dependency information.
 */
PUBLIC int sendServerHeader(int *extra, int noCompress)
{
  int n, rc = 0;
  z_stream *c_stream = NULL;	/* Compression stream */

  if (standalone && (state->loggedOutput == 1)&& (ServerRecovering != READLOG))
    stableLogOutput();
 
  state->Header[0] = '\0';

  if (!state->openConnection) {
    if (state->first) {
#ifdef NPH
      sprintf(state->Header, "HTTP/1.0 %d %s%c%c%s%c%c", state->status,
	      StatusMsg(state->status), CR, LF,
	      (standalone ? "Server: Standalone Rover/$Revision: 1.118 $" :
	       "Server: NCSA/1.4.1+Rover/$Revision: 1.118 $"), CR, LF);
#endif /* NPH */
      sprintf(line, "Content-type: application/x-object%c%c%c%c", CR, LF, CR, LF);
      strcat(state->Header, line);
    }
  } else {
    /*
     * Generate a header for the client's httpd
     *		ROVER <url> <session> <sessionID> [<logID>]
     */
    sprintf(state->Header, "ROVER %s %s %d %d\n", state->Object,
	    state->Service, state->ServiceID, state->LogID);

    /* Open a connection to the client here */
    if (!state->httpdHost || !state->port) {
      fprintf(stderr, "Missing httpdHost (%s) or port (%d) information\n",
	      state->httpdHost, state->port);
      return(0);
    }
    fprintf(stderr, "Opening HTTP connection to the client\n");
    if ((state->tcp->sock = openTCP(state->httpdHost, state->port)) == -1) {
      fprintf(stderr,"openTCP to %s:%d failed\n",state->httpdHost,state->port);
      exit(-1);
    }
    state->Stdin = fdopen(state->tcp->sock, "r");/* Convert to streams. */
    state->Stdout = fdopen(state->tcp->sock, "w");
    if (!state->Stdin || !state->Stdout) {
      fprintf(stderr, "fdopen(state->tcp->sock) failed");
      exit(-1);
    }
  }

  sprintf(line, "%2d %6d %7d %9d %1d %1d %1d %7d %4d\n",
	  state->ObjType, state->LogID, state->CodeSize, state->DataSize,
	  state->keepalive, state->compress, state->Cacheable, state->LogSize,
	  state->DVSize);
  strcat(state->Header, line);

  if (!state->OutSize && !state->DVSize) {
    fflush(state->Stdout);
    return(rc);
  }

  if (!noCompress && state->compress) {
    if (!(c_stream = InitCompress(state->compress))) {
      fprintf(stderr, "InitCompress error");
      exit(-1);
    }

    n = 0;
    if (Compress(c_stream, state->Stdout, state->output->buffer,state->OutSize,
		 output,2048,&n) != Z_OK) {
      fprintf(stderr, "Compress error");
      exit(-1);
    }
    rc = n;
    if (Compress(c_stream, state->Stdout, state->DVbuf, state->DVSize, output,
		 2048, &n) != Z_OK) {
      fprintf(stderr, "Compress error");
      exit(-1);
    }
    rc += n;
    *extra = 0;

    if (EndCompress(c_stream, state->Stdout, output, 2048, &n) != Z_OK) {
      fprintf(stderr, "EndCompress error");
      exit(-1);
    }
    rc += n;
  } else {
    rc = fwrite(state->output->buffer, sizeof(char), state->OutSize,
		state->Stdout);
    *extra = fwrite(state->DVbuf, sizeof(char), state->DVSize, state->Stdout);
  }

  fflush(state->Stdout);
  return(rc);
}
  fprintf(stderr, "HTTP/1.0 %d %s%c%c", state->status,
	  StatusMsg(state->status), CR, LF);
  fprintf(stderr, "Server: Standalone Rover/$Revision: 1.118 $\n");
  fprintf(stderr, "Server: NCSA/1.3+Rover/$Revision: 1.118 $\n");
  fprintf(stderr, "Content-type: application/x-object\n\n");
  fprintf(stderr, "%2d %6d %7d %9d %1d %1d %7d %4d\n", state->ObjType, LogID,
	  state->CodeSize, state->DataSize, state->compress, state->Cacheable,
	  state->LogSize, state->DVSize);

  fwrite(state->output->buffer, sizeof(char), state->OutSize, stderr);
  fwrite(state->DVbuf, sizeof(char), state->DVSize, stderr);
#endif
