 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: rover.h
 *
 * Description: Rover server-side routines.
 *
 *
 * Global Functions: None
 * Global Variables: None
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#ifndef HTROV_H
#define HTROV_H

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <sys/stat.h>
#include <pwd.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <setjmp.h>

#ifdef linux
#include <stdarg.h>
#include <unistd.h>
#else
#include <errno.h>
#include <varargs.h>
#endif /* linux */
#include <tcl.h>
#include "RoverTypes.h"

#define PUBLIC
#define PRIVATE static

#include "database.h"

#define SLOW 0
#define FAST 1

#define ROVER_ERROR 	-1
#define ROVER_FATAL 	-2
#define ROVER_FORK	-3
#define ROVER_PIPE	-4
#define ROVER_SIGNAL	-5

extern double TotalServerTime, TotalAppTime;
extern int TotalTimeCount;

typedef enum _CMDtype {
  EXPORTCMD,
  IMPORTCMD,
  IMPORTDATA,
  IMPORTBINARY,
  HTTP,
  TEST,
  RESET,
  SMTP,
  NONE
} CMDtype;

typedef enum _ServerRecoverype {
  NORMALOP,
  READLOG,
  REDO
} ServerRecoverType;

extern ServerRecoverType ServerRecovering;

typedef enum _Codes {
  OK = 200,			/* OK */
  CREATED =  201,		/* Created */
  ACCEPTED = 202,		/* Accepted */
  PROVIS_INFO = 203,		/* Provisional Information */
  NO_RESPONSE = 204,		/* No Response */
  DELETED = 205,		/* Deleted */
  MODIFIED = 206,		/* Modified */
  MOVED_PERM = 301,		/* Moved Permanently */
  MOVED_TEMP = 302,		/* Moved Temporarily */
  METHOD = 303,			/* Method */
  NOT_MODIFIED = 304,		/* Not Modified */
  BAD_REQUEST = 400,		/* Bad Request */
  UNAUTHORIZED = 401,		/* Unauthorized */
  PAYMENT_REQUIRED = 402,	/* Payment Required */
  FORBIDDEN = 403,		/* Forbidden */
  NOT_FOUND = 404,		/* Not Found */
  METHOD_NOT_ALLOW = 405,	/* Method Not Allowed */
  NONE_ACCEPTABLE = 406,	/* None Acceptable */
  PROXY_AUTH_REQ = 407,		/* Proxy Authentication Required */
  REQUEST_TIMEOUT = 408,	/* Request Timeout */
  INTERNAL_ERROR = 500,		/* Internal Error */
  NOT_IMPLEMENTED = 501,	/* Not Implemented */
  BAD_GATEWAY = 502,		/* Bad Gateway */
  SERVICE_UNAVAIL = 503,	/* Service Unavailable */
  GATEWAY_TIMEOUT = 504		/* Gateway Timeout */
} Codes;

typedef enum _LOGtype {
  ROVER_LOG_ERROR = -1,
  ROVER_LOG_VERSION = 0,
  ROVER_LOG_START,
  ROVER_LOG_FORK,
  ROVER_LOG_RESULT,
  ROVER_LOG_RECOVER,
  ROVER_LOG_STABLE_VAR,
  ROVER_LOG_STABLE_DCL,
  ROVER_LOG_STABLE_HTDCL,
  ROVER_LOG_STABLE_HTVAR,
  ROVER_LOG_DATA,
  ROVER_LOG_OUTPUT,
  ROVER_LOG_MSGID
} LOGtype;

#define ROVER_SERVER_LOG_VERSION	4
#define ROVER_SERVER_LOG_DIR		"/rover-log"

#define MAXSTATES	512

extern double elapsedTime(struct timeval tv1, struct timeval tv2);
extern char *copyString(char *string);
extern int sendServerResponse(int *extra, int noCompress);
extern void generateServerHeader();
extern void mailsetup();
extern CMDtype Name2Cmd(char *name);
extern char *rover_ctime(time_t *stamp);
extern void tcl_importdata(char *object);
extern char *StatusMsg(Codes code);
extern void errorExit(char *msg);
extern void fatalError(char *msg);

extern int stableHashInit(char *tableName);
extern char *stableHashFetch(char *tableName, char *entryName);
extern int stableHashSetEntry(char *tableName, char *entryName, char *data,
			      int size, char **old);
extern void stableLogOutput();

extern void outstrInit(int size);
extern void outstrResize(int size);
extern void emitData();
extern void emitLog();
extern void outstrL(time_t timestamp, char *host, int logID, RoverOpStatus status,
		    char *operation);
extern void outstrD(char *string);
extern void outstrC(char *string);
extern void outstrBin(void *bin, int count);
extern void outstrnD(char *string, int count);
extern void outstrnC(char *string, int count);
extern void outstrVD(char *, int count, ...);
extern void outstrVC(char *, int count, ...);

extern void deleteTclInterp();
extern void testExit(int size);
extern void set_serverID();

extern void RoverRootDirsInit();
extern void sendHeader(FILE *sock, char *url, int compress, int image,
		       int logID, char *source, char *path,
		       unsigned long time1, unsigned long time2,
		       unsigned long time3, int visible, int prefetch);
extern void sendMailHeader(FILE *file, char *url, char *origUrl,
			   int docompress, int image, char *rhost, int rport,
			   int logID, char *source, char *path, char *user,
			   unsigned long time1, unsigned long time2,
			   int prefetch);
extern void LogChange(char *action);

/* Definitions for procedures not in the include files */
extern char **environ;
extern int fflush(FILE *stream);
extern int pclose(FILE *stream);
extern int fclose(FILE *stream);

#ifndef linux
extern int setreuid(int ruid, int euid);
extern char *crypt(char *key, char *salt);
#endif /* !linux */

#ifdef sparc
#ifndef __cplusplus
extern int putenv(char *string);
extern int system(char *string);
extern int fread (char *ptr, int size, int nitems, FILE *stream);
extern int fputc(char c, FILE *stream);
extern int fputs(char *s, FILE *stream);
extern int printf(char *format, ... );
extern int fprintf(FILE *stream, char *format, ...);
extern int sscanf(char *s, char *format, ...);
extern int fwrite (char *ptr, int size, int nitems, FILE *stream);
extern char *index(char *s, char c);
extern int gettimeofday(struct timeval *tp, struct timezone (*tzp));
#endif
#endif /* sparc */

#ifndef INITSIZE
#define INITSIZE 	8192
#endif /* INITSIZE */
#define NPH

extern FILE *ServerLogFile;
extern FILE *ServerLogFile2;
extern Tcl_HashTable Info;
extern Tcl_HashEntry *hentry;
extern int RoverDebug, ServerLogFD, ServerLogFD2, ServerHardFlush;
extern unsigned int ServerID;
extern int NullTrace, ServerNoLog,  standalone;
extern char *mhdir, maildir[256], *ServerLogDir;
extern int ServerUseStableVars;
extern int ServerUseHash;
extern int ServerIgnoreHash;

extern int logClean;

extern uid_t RoverUid;
extern gid_t RoverGid;
extern int RoverRootID;
extern char *RoverRoot[];
extern char *RoverHomeEnv;
extern char *RoverUserEnv;

extern Tcl_Interp *globalTclInterp;
extern struct timeval StartTv;
extern struct timezone Tz;
extern struct timeval ServerStartTv, ServerEndTv;
extern struct timeval AppStartTv, AppEndTv;

#endif /* HTROV_H */
