 /**********************************************************************
 *
 *
 *		            Rover Toolkit
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: roverhttp.c
 *
 * Description: Rover Server HTTP module.
 *		Originally server gateway program for Rover Mosaic.
 *
 *
 * Global Functions: None
 * Global Variables: None
 *
 * 
 *
 ***********************************************************************
 *   Copyright 1995-1998
 *   Anthony D. Joseph and Alan deLespinasse
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include "rover.h"
#include "rovertcp.h"
#include "roverhttp.h"
#include "RoverTCP.h"
#include "RoverZlib.h"
#include "RoverEscape.h"
#include "html.h"

PUBLIC int ServerUseHash = 1;
PUBLIC int ServerIgnoreHash = 0;

typedef struct _docInfo {
  time_t time;		/* Time document was loaded */
  char serverTime[48];	/* Time from server */
  int size;		/* Size of document */
} docInfo;

/* These structures implement a data structure built when HTML is parsed
   for other URLs to prefetch. This first one is for a simple linked list
   of URLs to fetch. */
typedef struct prefetchpage {
  char *url;
  int image;					/* Image? */
  struct prefetchpage* next;
  int prefetch;
} prefetchpage;

/* And this one is the header for the above list. It's passed to the parsing
   callback functions (as the extra void* in the HTML library-- see html.h)
   so they know where to add the new list member, how to interpret relative
   URLs, etc. */
typedef struct parsedata {
  ServerState *state;				/* Server state variable */
  int prefetch;					/* Remaining prefetch depth */
  char *source;					/* Page being fetched */
  char *sourcetitle;				/* Page being fetched */
  char *base;					/* URL of same */
  struct prefetchpage *pagetop, *pagebottom; 	/* Prefetch linked list */
  int intitle;		/* Flag indicating HTML parser is in document title */
} parsedata;

/* Defined below (we've got some recursion going on here). */
PRIVATE char *fetch(ServerState *s, URLinfo *uinfo, char *method, int image,
		    char *pagetitle, char *source, char *sourcetitle,
		    char *path, int prefetch);

/* Release storage */
PRIVATE void cleanup(ServerState *s)
{
  HTTPInfo *http = (HTTPInfo *)s->plugInData;

  /* ### This should free prefetch info, parsedata->base, and the parser */
  if (!http) return;
  if (http->pagetitle) free(http->pagetitle);
  if (http->url) free(http->url);
  if (http->rhost) free(http->rhost);
  if (http->source) free(http->source);
  if (http->sourcetitle) free(http->sourcetitle);
  if (http->path) free(http->path);
  if (http->user) free(http->user);
  if (http->reply) free(http->reply);
  if (http->mailcmd) free(http->mailcmd);
  if (http->wwwcmd) free(http->wwwcmd);
  free(http);
}

/* Dump everything we know to the error log and exit. */
PRIVATE void puke(ServerState *s, char *msg)
{
  HTTPInfo *http = (HTTPInfo *)s->plugInData;

  fprintf(stderr, ctime(&(http->respTime)));
  fprintf(stderr, "Fatal error: %s\n", msg);
  fprintf(stderr, "Hostname: %s\n", http->rhost);
  fprintf(stderr, "Port: %d\n", http->rport);
  fprintf(stderr, "URL: %s\n", http->url);
  fprintf(stderr, "Log ID: %d\n", http->logID);
  fprintf(stderr, "Pagetitle: %s\n", http->pagetitle);
  fprintf(stderr, "Source: %s\n", http->source);
  fprintf(stderr, "Sourcetitle: %s\n", http->sourcetitle);
  fprintf(stderr, "Path: %s\n", http->path);
  fprintf(stderr, "User: %s\n", http->user);
  fprintf(stderr, "Time1 (reqTime): %lu\n", (unsigned long) http->reqTime);
  fprintf(stderr, "Time2 (respTime): %lu\n", (unsigned long) http->respTime);
  fprintf(stderr, "Method: %s\n", http->method);
  fprintf(stderr, "Prefetch: %d\n", http->prefetch);
  fprintf(stderr, "Preftime: %ld\n", http->preftime);
  fprintf(stderr, "Compress: %d\n", http->docompress);
  fprintf(stderr, "Inline: %s\n", http->doinline ? "yes" : "no");
  fprintf(stderr, "MIME header: %s\n", http->MIMEHdr);
  fprintf(stderr, "Reply: %s\n", http->reply);
  fprintf(stderr, "Mail command: %s\n", http->mailcmd);
  fprintf(stderr, "WWW command: %s\n", http->wwwcmd);
  fprintf(stderr, "Remainder of stdin:\n");
  fputs(http->pointer, stderr);
  fprintf(stderr,"\n\n");
  fflush(stderr);
  s->status = 500;

  /* Release storage space */
  cleanup(s);
  errorExit("HTTP proxy error");
  exit(0);
}

/* Dump everything we know to the error log. */
PRIVATE void info(ServerState *s, char *msg)
{
  HTTPInfo *http = (HTTPInfo *)s->plugInData;

  fprintf(stderr, ctime(&(http->respTime)));
  fprintf(stderr, "Info: %s\n", msg);
  fprintf(stderr, "Hostname: %s\n", http->rhost);
  fprintf(stderr, "Port: %d\n", http->rport);
  fprintf(stderr, "URL: %s\n", http->url);
  fprintf(stderr, "Log ID: %d\n", http->logID);
  fprintf(stderr, "Pagetitle: %s\n", http->pagetitle);
  fprintf(stderr, "Source: %s\n", http->source);
  fprintf(stderr, "Sourcetitle: %s\n", http->sourcetitle);
  fprintf(stderr, "Path: %s\n", http->path);
  fprintf(stderr, "User: %s\n", http->user);
  fprintf(stderr, "Time1 (reqTime): %lu\n", (unsigned long) http->reqTime);
  fprintf(stderr, "Time2 (respTime): %lu\n", (unsigned long) http->respTime);
  fprintf(stderr, "Method: %s\n", http->method);
  fprintf(stderr, "Prefetch: %d\n", http->prefetch);
  fprintf(stderr, "Preftime: %ld\n", http->preftime);
  fprintf(stderr, "Compress: %d\n", http->docompress);
  fprintf(stderr, "Inline: %s\n", http->doinline ? "yes" : "no");
  fprintf(stderr, "MIME header: %s\n", http->MIMEHdr);
  fprintf(stderr, "Reply: %s\n", http->reply);
  fprintf(stderr, "Mail command: %s\n", http->mailcmd);
  fprintf(stderr, "WWW command: %s\n", http->wwwcmd);
  fprintf(stderr, "Remainder of stdin:\n");
  fputs(http->pointer, stderr);
  fprintf(stderr, "\n\n");
  fflush(stderr);
}

/* Plain text callback for HTML parser. Sets the pagetitle global var if
   it's between a <title> and a </title>, as determined by the next
   function. */
void textcallback(struct HTMLparser *p)
{
  struct parsedata *pd = (struct parsedata *) p->data;
  HTTPInfo *http = (HTTPInfo *)(pd->state->plugInData);
  char *t;

  if ((!pd->intitle) || http->pagetitle) return;
fprintf(stderr, "Title callback %s\n", p->buf);
  t = &(p->buf[strspn(p->buf, " \t\r\n")]);
  if (!*t) return;
  t = strtok(t, "\n\r");
  if (!t || !*t) return;
  http->pagetitle = copyString(t);
fprintf(stderr, "Title is: %s\n", t);
}

/* Add a url to the prefetch list for the parent page. Adds the url if the
   prefetch depth has not been exceeded and the url has not already been 
   added to the prefetch list (doesn't cannonicalize the name; so some 
   duplicates may be missed - but will be caught by fetch's stable hash
   table).
 */
PRIVATE void prefetchPage(char *url, struct parsedata *pd, int now, int image)
{
  struct prefetchpage *pfp;
  URLinfo uinfo;
  char *errString;

  if ((pd->prefetch < 0) || ((pd->prefetch == 0) && !image)) return;
  fprintf(stderr, "Prefetching %s%s (%s) [%d]\n", (image ? "image " : ""), url,
	  (now ? "now" : "delayed"), pd->prefetch);
  if (!now) {
    /* Check for duplicates */
    pfp = pd->pagetop;
    while (pfp) {
      if (!strcmp(pfp->url, url)) {
	fprintf(stderr, "Skipping %s, already in prefetch queue\n", url);
	return;
      }
      pfp = pfp->next;
    }
    pfp = (struct prefetchpage *) malloc(sizeof(struct prefetchpage));
    pfp->url = copyString(url);
    pfp->image = image;
    pfp->prefetch = pd->prefetch;
    pfp->next = NULL;			/* Add the URL to the ... */
    if (pd->pagetop) {			/*   tail of the list */
      pd->pagebottom->next = pfp;
      pd->pagebottom = pfp;
    } else				/*   head of the list */
      pd->pagetop = pd->pagebottom = pfp;
  } else {
    uinfo.url = copyString(url);
    uinfo.base = pd->base;
    if ((errString = convertURL(&uinfo)) != NULL) puke(pd->state, errString);
    fetch(pd->state,&uinfo,"GET",image,NULL,pd->base,NULL, NULL, pd->prefetch);
    free(uinfo.url);
  }
}

/* The callback function for parsing HTML. Goes and fetches inlined images
   right away, so they hopefully get there before the main document.
   Documents to prefetch are stored in a list for later fetching. Does
   not handle image maps correctly. */
void tagcallback(struct HTMLparser *p)
{
  int found = 0;
  struct parsedata *pd = (struct parsedata *) p->data;
  HTTPInfo *http = (HTTPInfo *)(pd->state->plugInData);
  char *t = strtok(p->buf, " <>=\t\n\r");
  if (!strcasecmp(t, "a")) {		/* it's a hypertext link */
    if (pd->prefetch <= 0) return; 	/* Reached limit of recursion */
    while (1) {
      t = strtok(NULL, " <>=\t\n\r");	/* find the HREF */
      if (!t) return;
      if (strcasecmp(t, "href") == 0) {
	found = 1; 
	break;
      }
    }
    if (!found) return;
    t = strtok(NULL, " <>=\"\t\n\r");	/* find the URL */
    if (!t) return;
    if (!strncasecmp("mailto:", t, 7) || !strncasecmp("news:", t, 5) ||
	!strncasecmp("javascript:", t, 7)) return; /* Ignore some hrefs */
    prefetchPage(t, pd, 0, 0);
  } else if (!strcasecmp(t, "img") || !strcasecmp(t, "image") ||
	     !strcasecmp(t, "fig")) {	/* it's an inlined image */
    if (!http->doinline) return;
    if (pd->prefetch < 0) return; 	/* Reached limit of recursion */
    while (1) {
      t = strtok(NULL, " <>=\t\n\r");	/* find the SRC reference */
      if (!t) return;
      if (strcasecmp(t, "src") == 0) {
	found = 1;
	break;
      }
    }
    if (!found) return;
    t = strtok(NULL, " <>=\"\t\n\r");	/* find the URL */
    if (!t) return;
    prefetchPage(t, pd, 0, 1);
  } else if (!strcasecmp(t, "body")) {	/* check for background image */
    if (!http->doinline) return;
    if (pd->prefetch < 0) return; 	/* Reached limit of recursion */
    while (1) {
      t = strtok(NULL, " <>=\t\n\r");	/* find the SRC reference */
      if (!t) return;
      if (strcasecmp(t, "background") == 0) {
	found = 1;
	break;
      }
    }
    if (!found) return;
    t = strtok(NULL, " <>=\"\t\n\r");	/* find the URL */
    if (!t) return;
    prefetchPage(t, pd, 0, 1);
  } else if (!strcasecmp(t, "title")) {
    pd->intitle = 1;
  } else if (!strcasecmp(t, "/title")) {
    pd->intitle = 0;
  }
}

/* Initialize prefetching and parser */
PRIVATE inline struct HTMLparser *prefetchInit(ServerState *s,
					       struct parsedata *pd,
					       int prefetch, char *source,
					       char *sourcetitle, char *base)
{
  pd->state = s;
  pd->intitle = 0;
  pd->prefetch = prefetch;
  pd->source = source;
  pd->sourcetitle = sourcetitle;
  pd->base = base;
  pd->pagetop = pd->pagebottom = NULL;
  return(HTMLnewparse(textcallback, tagcallback, (void *) pd));
}

PRIVATE int roverHTTPOpen(ServerState *s, HTTPInfo *http, char *url)
{
  /* Open the connection to the browser interface application,
   *   if necessary. A connection is opened only for the second or
   *   later objects and when compression isn't being used. 
   */
  int opened = 0;

  if (!http->docompress) {
    if (!http->first) {
      opened = 1;
      http->osock = openTCP(http->rhost, http->rport);
      if (http->osock == -1) puke(s, "openTCP to browser failed.");
      
      /* Convert to streams. */
      http->osockout = fdopen(http->osock,"w"); 
      if (!http->osockout) puke(s, "fdopen(http->osock) failed.");
    }
  } else {
    http->osock = s->tcp->sock;
    http->osockout = s->Stdout;
  }
  if (http->first || opened)
    fprintf(http->osockout, "CACHE %s HTTP/1.0\n", url);	/* update */

  return(opened);
}

/* Fetch a document using HTTP, and don't give up easily. Send it back
   using mailcmd. Parse HTML, if necessary, and do prefetching stuff.
   Returns an error message string; NULL on success. */
PRIVATE char *fetch_http(ServerState *st, URLinfo *uinfo, char *method,
			 int image, char *pagetitle, char *source,
			 char *sourcetitle, char *path, int prefetch)
{
  FILE *sockin, *sockout;
  struct parsedata pd;
  struct prefetchpage *pfp, *next;
  int sock, n, linebeg, respcode, html = 0, redirect = 0, i, size = 0;
  char s[1024], c[2048], *line, *type = NULL, *dest = NULL, *tmp, date[48];
  char *url = uinfo->url;
  URLinfo uinfo2;
  struct HTMLparser *parser = NULL;
  z_stream *c_stream = NULL;	/* Compression stream */
  docInfo *info;
  HTTPInfo *http = (HTTPInfo *)st->plugInData;
  int opened = 0, parseInit = 0, first = http->first;

  *date = '\0';
  fprintf(stderr, "Fetching %s%s [%d]\n", (image ? "image ": ""),url,prefetch);
  if ((sock=openTCP(uinfo->hostname,uinfo->port))==-1) return "openTCP failed";
  sockin = fdopen(sock,"r");	/* Convert to streams. */
  sockout = fdopen(sock,"w");
  if (!sockin || !sockout) return "fdopen(sock) failed.";

  fprintf(sockout, "%s %s HTTP/1.0\n", method, uinfo->urlpath); /* request */
  if (RoverDebug) fprintf(stderr, "%s %s HTTP/1.0\n", method, uinfo->urlpath); 

  fprintf(sockout, "Host: %s\n", uinfo->hostname);
  if (http->MIMEHdr) {
    if (RoverDebug) fprintf(stderr,"Got browser MIME hdr: %s\n",http->MIMEHdr);
    fwrite(http->MIMEHdr, 1, strlen(http->MIMEHdr), sockout);
  } else {
    if (RoverDebug) fprintf(stderr, "No browser MIME hdr\n");
    putc('\n', sockout);
  }

  if (strlen(http->pointer) != strspn(http->pointer," \t\r\n")) {
    if (RoverDebug) fprintf(stderr, "Got browser data: `%s'\n", http->pointer);
    fwrite(http->pointer, 1, strlen(http->pointer), sockout);
  } else {
    if (RoverDebug) fprintf(stderr, "No browser data\n");
    putc('\n', sockout);
  }
  http->pointer = "";

  fflush(sockout);

  fgets(s,1000,sockin);
  n = sscanf(s,"%*s %d", &respcode);
  if (!n) respcode = 0;		/* For obsolete 0.9 servers */
  if (RoverDebug) fprintf(stderr, "Response is `%s', code %d\n", s, respcode);
  if ((respcode>300) && (respcode<400)) { /* been redirected */
    if (RoverDebug) fprintf(stderr, "Got HTTP redirect: %d\n", respcode);
    redirect = 1;
    /* Mark the page not visible (the redirected page will be visible) */
    http->visible = 0;
  }

  opened = roverHTTPOpen(st, http, url);
  http->logID = -1;
  pd.base = NULL;

  if (!path)
    http->visible = 0; /* !path? ==> not explicitly requested, hide it*/

  tmp = strrchr(url,'.');	/* Find last . and check for HTML extension */
  if (tmp && (!strcasecmp(tmp,".html") || !strcasecmp(tmp,".shtml") ||
	      !strcasecmp(tmp,".htm") || !strcasecmp(tmp,".shtm"))) {
    /* Parse HTML */
    parseInit = html = 1;
    parser = prefetchInit(st,&pd,prefetch,source,sourcetitle, copyString(url));
    if (RoverDebug) fprintf(stderr, "Parser initialized\n");
  }

  /* Send the header info. */
  sendHeader(http->osockout, url, http->docompress, image, http->logID,
	     source, path, http->reqTime, http->respTime, 0, http->visible,
	     prefetch);

  /* Write the server's response line */
  if (http->docompress) {
    if (!(c_stream = InitCompress(http->docompress)))
      puke(st, "InitCompress error");
    if (Compress(c_stream, http->osockout, s, strlen(s), c, 2048, &i) != Z_OK)
      puke(st, "Compress error");
  } else fwrite(s, 1, strlen(s), http->osockout);

  if (!strncasecmp(url, "http:",5) || redirect) {
    /* Only HTTP or redirects use MIME headers */
    for (;!feof(sockin);) {
      fgets(s,1000,sockin);
      if (http->docompress) {
	if (Compress(c_stream, http->osockout, s, strlen(s),c,2048,&i) != Z_OK)
	  puke(st, "Compress error");
      } else fwrite(s, 1, strlen(s), http->osockout);
      linebeg = strspn(s, " \t\r\n"); /* and parse for content-type */
      if (linebeg==strlen(s)) break;
      line=&s[linebeg];
      if (strncasecmp(line,"content-type:",13)==0)
	type=copyString(strtok(&line[13]," \t\n\r"));
      if (!strncasecmp(line,"date:",5)) strcpy(date,strtok(&line[5],"\t\n\r"));
      if (!strncasecmp(line,"last-modified:",14))
	strcpy(date,strtok(&line[14],"\t\n\r"));
      if (!redirect) continue;

      if (!strncasecmp(line,"location:",9)) dest=strtok(&s[9]," \t\n\r");
      else if (!strncasecmp(line,"url:",4)) dest=strtok(&s[4]," <>\t\n\r");
      if (dest) {
	/* Prefetch page instead of fetching it inline */
	if (RoverDebug) fprintf(stderr, "Redirect to %s\n", dest);
	if (!parseInit) {
	  parseInit = html = 1; 
	  parser = prefetchInit(st, &pd, prefetch, source, sourcetitle,
				copyString(source));
	} else {	/* Change the base URL to be the current source URL */
	  if (pd.base) {
	    free(pd.base);
	    pd.base = NULL;
	  }
	  if (source) pd.base = copyString(source);
	}
	pd.prefetch++;	/* Bump up the prefetch depth so that the redirected
			   page will be fetched the same as this page */
	prefetchPage(dest, &pd, 0, 0);
	pd.prefetch--;
	redirect = 0;
      }
    }
  }

  if (html) pd.prefetch = prefetch;

  /* This should probably also check for "application/html", etc. */
  if (!parseInit && !html && ((prefetch >= 0) || http->doinline) && type &&
      !strcasecmp(type,"text/html")) {
    parseInit = html = 1; 
    parser = prefetchInit(st,&pd,prefetch,source,sourcetitle, copyString(url));
  }
  if (type) free(type);

  n = 0;
  while (!feof(sockin) && ((n = fread(s, 1, 1000, sockin)) != 0 )) {
    /* Do the whole message body */
    size += n;
    if (http->docompress) {
      if (Compress(c_stream, http->osockout, s, n, c, 2048, &i) != Z_OK)
	puke(st, "Compress error");
    } else fwrite(s, 1, n, http->osockout);
    if (html) HTMLparse(parser,s,n);
  }
  if (html) HTMLendparse(parser);

  if (http->docompress) {
    if (EndCompress(c_stream, http->osockout, c, 2048, &i) != Z_OK)
      puke(st, "EndCompress error");
  }
/*  fflush(http->osockout); */

  if (!http->docompress) {
    fclose(http->osockout);
    close(http->osock);
    opened = 0;
  }

  fclose(sockin);
  fclose(sockout);
  close(sock);

  
  if (ServerUseHash) {
    char hashTableName[256]; /* Hash table name for HTTP info */
    sprintf(hashTableName, "HTTP_Table_%s@%s", st->ClientUser, st->rhost);
    if (!(info = (docInfo *) stableHashFetch(hashTableName, url)))
      fprintf(stderr, "Unable to update size for url: %s\n", url);
    else {
      info->size = size;
      strcpy(info->serverTime, date);
    }
  }

  fprintf(stderr, "Done sending page `%s' [fetch_http]\n", url);
  http->first = 0;

  if (html && (prefetch >= 0)) {
    for (pfp = pd.pagetop; pfp;) {
      fprintf(stderr,"Sending prefetch page %s (base: %s)\n",pfp->url,pd.base);
      uinfo2.url = pfp->url;
      uinfo2.base = pd.base;
/*      uinfo2.base = url; */
      if ((tmp = convertURL(&uinfo2)) != NULL) puke(st, tmp);
      pfp->url = uinfo2.url;
      fetch(st, &uinfo2, "GET", pfp->image, NULL, url,NULL,NULL,pfp->prefetch);
      next = pfp->next;
      free(pfp->url);
      free(pfp);
      pfp = next;
    }
  }

  if (pd.base) free(pd.base);
  if (first) fprintf(http->osockout, "end: \n");
  if (opened) {
    fclose(http->osockout);
    close(http->osock);
  }
  return NULL;
}

/* Fetch a document by calling the line-mode browser, www. Could work
   for non-form HTTP fetches, but the above function is more efficient
   and does all the extra prefetching stuff that this one doesn't, and
   can also do forms. This function is only used for non-HTTP fetches.
   Returns an error message string; NULL on success. */
PRIVATE char *fetch_other(ServerState *st, char *url, char *pagetitle,
			  char *source, int image, char *sourcetitle,
			  char *path)
{
  FILE *wwwin;
  char s[1024], c[2048];
  int i, size = 0;
  size_t n;
  z_stream *c_stream = NULL;	/* Compression stream */
  char *wwwcmd;			/* command line for line-mode browser */
  docInfo *info;
  HTTPInfo *http = (HTTPInfo *)st->plugInData;
  int opened = 0, first = http->first;

  /* Call the line-mode browser. */
  wwwcmd = malloc(strlen(url)+strlen(http->wwwcmd)+2);
  sprintf(wwwcmd, "%s %s", http->wwwcmd, url);
  fprintf(stderr, "WWW command `%s'\n", wwwcmd);

  if (!(wwwin = popen(wwwcmd, "r"))) return "popen(wwwcmd) failed.";

  opened = roverHTTPOpen(st, http, url);
  http->logID = -1;

  if (!path) 
    http->visible = 0; /* !path? ==> not explicitly requested, hide it */

  /* Send the header info. */
  sendHeader(http->osockout, url, http->docompress, image, http->logID,
	     source, path, http->reqTime, http->respTime,0,http->visible,0);

  if (http->docompress)
    if (!(c_stream = InitCompress(http->docompress)))
      puke(st, "InitCompress error");

  /* Now send the data */
  n = 0;
  while (!feof(wwwin) && ((n = fread(s,1,1000,wwwin)) != 0))
    size += n;
    if (http->docompress) {
      if (Compress(c_stream, http->osockout, s, n, c, 2048, &i) != Z_OK)
	puke(st, "Compress error");
    } else fwrite(s, 1, n, http->osockout);

  if (http->docompress) {
    if (EndCompress(c_stream, http->osockout, c, 2048, &i) != Z_OK)
      puke(st, "EndCompress error");
  }
  fflush(http->osockout);

  if (first) fprintf(http->osockout, "end: \n");
  if (opened) {
    fclose(http->osockout);
    close(http->osock);
  }

  pclose(wwwin);
  free(wwwcmd);

  if (ServerUseHash) {
    char hashTableName[256]; /* Hash table name for HTTP info */
    sprintf(hashTableName, "HTTP_Table_%s@%s", st->ClientUser, st->rhost);
    if (!(info = (docInfo *) stableHashFetch(hashTableName, url)))
      fprintf(stderr, "Unable to update size for url: %s\n", url);
    else {
      info->size = size;
      info->serverTime[0] = '\0';
    }
  }

  fprintf(stderr, "Done sending page `%s' [fetch_other]\n", url);
  http->first = 0;
  return NULL;
}

/* Fetch a URL. If the URL is relative, turns it into absolute. Doesn't
   fetch if the same URL has already been fetched. The stream pointer
   can optionally contain the rest of the request message after the
   first line; use NULL if it doesn't. */
PRIVATE char *fetch(ServerState *s, URLinfo *uinfo, char *method, int image,
		    char *pagetitle, char *source, char *sourcetitle,
		    char *path, int prefetch)
{
  char *old = NULL;
  docInfo *info;
  char *url = uinfo->url;
  HTTPInfo *http = (HTTPInfo *)s->plugInData;

  prefetch--;
  if (ServerUseHash) {
    char hashTableName[256]; /* Hash table name for HTTP info */
    sprintf(hashTableName, "HTTP_Table_%s@%s", s->ClientUser, s->rhost);
    if (!ServerIgnoreHash && stableHashFetch(hashTableName, url)) return(NULL);
    if ((info = (docInfo *)malloc(sizeof(docInfo))) == NULL) {
      fatalError("Server unable to allocate HTTP hash table entry");
    }
    info->time = http->reqTime;
    info->size = -1;
    info->serverTime[0] = '\0';
    if (stableHashSetEntry(hashTableName, url, (char *)info, sizeof(docInfo),
			   &old)) {
      if (!ServerIgnoreHash) {
	free(info);
	return NULL;
      }
    }
    if (old) free(old);
  }

  if (!strcasecmp(uinfo->protocol,"http"))
    return fetch_http(s, uinfo, method, image, pagetitle, source, sourcetitle,
		      path, prefetch);
  else
    return fetch_other(s, url, pagetitle, source, image, sourcetitle, path);
  free(url);
}

int resumeHTTP(ServerState *s)
{
  HTTPInfo *http = (HTTPInfo *)s->plugInData;

  /* Jump into the FSM. */
  while (1) {
    switch(http->state) {
    case HTTP_BEGIN:
      http->state = HTTP_NEED_CONNECTION;
      break;
    case HTTP_NEED_CONNECTION:

    case HTTP_NEED_REQUEST:
    case HTTP_SENT_REQUEST:
    case HTTP_NEED_BODY:
    case HTTP_GOT_BODY:
    case HTTP_CLEANUP:
    case HTTP_ABORT:
    default:
      fprintf(stderr, "Got unknown state %d [resumeHTTP:roverhttp.c]\n",
	      http->state);
    }
  }
  return 0;
}

int processHTTP(ServerState *s, int argc, char **argv, int SMTP)
{
  int rc, extra;
  char *www, *token, *errString;
  char hashTableName[256]; /* Hash table name for HTTP info */
  URLinfo uinfo;
  HTTPInfo *http;
  struct stat sbuf;

  if (s->plugInData) return(resumeHTTP(s));

  if ((s->plugInData = (void *)malloc(sizeof(HTTPInfo))) == NULL) {
    fatalError("Server unable to allocate HTTP info structure");
  }
  http = (HTTPInfo *)s->plugInData;

  /* Initialize variables on each pass */
  if (!SMTP) {
    RoverUnescape(s->code);
    if (RoverDebug) fprintf(stderr, "Got http command: %s\n", s->code);
    s->ObjType = ROVER_HTTP;
    rc = sendServerResponse(&extra, 0);
    s->CommandDone = 1;
  }

  http->state = HTTP_BEGIN;
  http->source = http->sourcetitle = http->path = NULL;
  http->MIMEHdr = http->user = http->pagetitle = http->method = NULL;
  http->wwwcmd = http->mailcmd = http->reply = http->url = http->rhost = NULL;
  http->rport = 0;
  http->prefetch = 0;
  http->first = http->doinline = http->visible = 1;
  http->reqTime = 0x7FFFFFFF;
  http->respTime = 0;
  http->preftime = 3600;
  http->respTime = time(NULL);
  http->pointer = s->code;
  http->logID = s->LogID;

  sprintf(hashTableName, "HTTP_Table_%s@%s", s->ClientUser, s->rhost);
  if (ServerUseHash) 
    stableHashInit(hashTableName);	/* Track already fetched URLs */

  /* Parse SMTP headers. */
  while (1) {
#ifndef sparc
    token = strsep(&(http->pointer), "\n\r");
#else
    /* Stupid sunos doesn't have strsep, so we implement it here */
    token = http->pointer;
    while (*(http->pointer) && (*(http->pointer) != '\n') &&
	   (*(http->pointer) != '\r'))
      http->pointer++;	
    if (!*(http->pointer))
      token = NULL;
    else {
      *(http->pointer) = '\0';
      http->pointer++;
    }
#endif
    if (!token) puke(s, "No header/body separator.");
    if (strspn(token, " \t\n\r") == strlen(token)) break;
    
    if (strncasecmp(token, "x-url:", 6) == 0)
      http->url = copyString(strtok(&token[6]," \t\n\r"));
    else if (strncasecmp(token, "x-reply: ", 9) == 0)
      http->reply = copyString(strtok(&token[9], " \t\n\r"));
    else if (strncasecmp(token, "x-hostname: ", 12) == 0)
      http->rhost = copyString(strtok(&token[12], " \t\n\r"));
    else if (strncasecmp(token,"x-port:",7)==0)
      http->rport = atoi(strtok(&token[7], " \t\n\r"));
    else if (strncasecmp(token,"x-logid:", 8) == 0)
      http->logID = atoi(strtok(&token[8], " \t\n\r"));
    else if (strncasecmp(token,"x-compress:",11)==0)
      http->docompress = atoi(strtok(&token[11], " \t\n\r"));
    else if (strncasecmp(token,"x-source:",9)==0)
      http->source = copyString(strtok(&token[9]," \t\n\r"));
    else if (strncasecmp(token, "x-sourcetitle: ", 15) == 0)
      http->sourcetitle = copyString(strtok(&token[15], "\n\r"));
    else if (strncasecmp(token, "x-path:", 7)==0)
      http->path = copyString(strtok(&token[7], " \t\n\r"));
    else if (strncasecmp(token, "x-user:", 7)== 0)
      http->user = copyString(strtok(&token[7], " \t\n\r"));
    else if (strncasecmp(token, "x-time1:", 8) == 0) {
      unsigned long ultime;
      sscanf(&token[8], "%lu", &ultime);
      http->reqTime = (time_t) ultime;
    }
    else if (strncasecmp(token, "x-prefetch:", 11) == 0)
      sscanf(&token[11],"%d", &(http->prefetch));
    else if (strncasecmp(token, "x-preftime:", 11) == 0)
      sscanf(&token[11], "%ld", &(http->preftime));
    else if (strncasecmp(token,"x-inline:",9)==0) {
      char *t = strtok(&token[9], " \t\n\r");
      http->doinline = (t[0]=='y' || t[0]=='Y') ? 1 : 0;
    } else if (strncasecmp(token,"x-request:",10)==0) {
      http->method = strtok(&token[10]," \t\n\r");	/* Get the method */
      if (!http->method) puke(s, "No method.");
      /* Now, get the URL if it wasn't in the X-url field */
      if (!http->url) {
	if (RoverDebug) fprintf(stderr, "No X-url, checking request line\n");
	http->url = strtok(NULL," \t\n\r");
      }
    } else if (strncasecmp(token,"x-mimehdr:",10)==0) {	/* Get the MIME hdr */
      http->MIMEHdr = copyString(RoverUnescape(strtok(&token[10]," \t\n\r")));
    }
  }
    
#if 0
  if (((long) http->respTime - (long) http->reqTime) <
      http->preftime)
    http->prefetch = 0;
#endif

  http->pointer++;	/* Skip \n */
  if (RoverDebug) fprintf(stderr, "Pointer is `%s'\n", http->pointer);
  if (!http->url) puke(s, "No URL.");
  if (RoverDebug) fprintf(stderr, "HTTP url is %s\n", http->url);

  /* Finally, get the MIME hdr if one was sent */
  if (!ServerIgnoreHash) {
    if (http->MIMEHdr) {
      /* Save the MIME header */
      www = NULL;
      rc =stableHashSetEntry(hashTableName,"MIME_Header",strdup(http->MIMEHdr),
			     strlen(http->MIMEHdr)+1, &www);
      if (www) free(www);
    } else
      http->MIMEHdr = stableHashFetch(hashTableName, "MIME_Header");
  }
  if (RoverDebug) fprintf(stderr, "MIME Header is `%s'\n", http->MIMEHdr);

  if (!http->rport) http->rport = s->port;
  if (!http->rhost) http->rhost = copyString(s->rhost);
  if (s->remoteName && strcmp(s->remoteName, "UNKNOWN_HOST") &&
      strcmp(http->rhost, s->remoteName))
    http->rhost = copyString(s->remoteName);
  if (RoverDebug) fprintf(stderr, "rhost is %s\n", http->rhost);

  http->SMTP = SMTP;
  if (SMTP) {
    /* Figure out how to send mail and call line-mode browser. */
    if (argc > 1) {
      http->mailcmd = malloc(strlen(argv[1])+strlen(http->reply)+2);
      sprintf(http->mailcmd, "%s %s", argv[1], http->reply);
    } else {
      http->mailcmd = malloc(strlen(http->reply)+10);
      sprintf(http->mailcmd, "/usr/lib/sendmail %s", http->reply);
    }

    if (argc > 2) {
      http->wwwcmd =  malloc(strlen(argv[2])+2);
      sprintf(http->wwwcmd, "%s ", argv[2]);
    } else {
      /* Locate the line-mode browser. */
      if (!stat("/usr/local/bin/www", &sbuf))
	www = "/usr/local/bin/www";
      else {
	www = "www";
	fprintf(stderr, "Warning! Server may not find W3 LineMode browser");
      }	
      http->wwwcmd = malloc(strlen(www)+strlen(" -n -source ")+1);
      sprintf(http->wwwcmd, "%s -n -source ", www);
    }
    if (RoverDebug) {
      fprintf(stderr, "Mail command `%s'\n", http->mailcmd);
      fprintf(stderr, "WWW command `%s'\n", http->wwwcmd);
    }
  }

  if (RoverDebug) info(s, "Scan complete");

  /* Get the document. */
  uinfo.url = http->url;
  uinfo.base = NULL;
  if ((errString = convertURL(&uinfo)) != NULL) puke(s, errString);
  if ((errString = fetch(s, &uinfo, http->method, 0, http->pagetitle,
			 http->source, http->sourcetitle, http->path,
			 http->prefetch)) != NULL)
    puke(s, errString);

  /* Release storage space */
  http->url = NULL;
  cleanup(s);
  return 0;
}
